/* SPDX-License-Identifier: GPL-2.0 */
/*
 * Copyright (c) 2018 Synopsys, Inc. and/or its affiliates.
 * Synopsys DesignWare eDMA controller driver.
 */

#ifndef __PEDD_H__
#define __PEDD_H__

#include "pedd_types.h"


/*******************************************************************************
 *
 * MACRO DEFINITIONS
 *
 ******************************************************************************/

#define DRIVER_VERSION          "1.0"
#define DRVER_NAME              "pedd"
#define DRVER_CLASS_NAME        "snps_pcie_ep"         
#define MAX_DEVICE              (1)
#define PEDD_CDEV_MAJOR         (0)
#define PEDD_NR_DEVS            (1)
#define BAR_MEM                 (0)
#define PEDD_BAR_NUM            (6)

#define DIRECTION_TO_DEVICE_WRITE	0	//Write data from user buffers to Device memory (DDR Buffer)
#define DIRECTION_TO_DEVICE_READ	1	//Read data from device into user buffers

#ifndef VM_RESERVED
	#define VMEM_FLAGS (VM_IO | VM_DONTEXPAND | VM_DONTDUMP)
#else
	#define VMEM_FLAGS (VM_IO | VM_RESERVED)
#endif

/**
 * The device id of PCIe RC device
 */
#define RC_SNPS_VENDOR_ID	0x1E7C
#define RC_SNPS_DEVICE_ID	0xBCA1
/**
 * Default data pattern.
 */
#define DEFAULT_PATTERN 0x00000001

/**
 * Get the source id from channel direction and ddr usage.
 */
#define GET_CHAN_SRC(dirc, ddr) (dirc == DMA_WRITE ? \
                (ddr ? DMA_CHAN_DDR : DMA_CHAN_LOC) : DMA_CHAN_HST)
/**
 * Get the destination id from channel direction and ddr usage.
 */
#define GET_CHAN_DST(dirc, ddr) (dirc == DMA_READ ? \
                (ddr ? DMA_CHAN_DDR : DMA_CHAN_LOC) : DMA_CHAN_HST)

/* obtain the 32 most significant (high) bits of a 32-bit or 64-bit address */
#define pci_dma_h(addr) ((addr >> 16) >> 16)
/* obtain the 32 least significant (low) bits of a 32-bit or 64-bit address */
#define pci_dma_l(addr) (addr & 0xffffffffUL)


/*******************************************************************************
 *
 * DATA STRUCTURE DEFINITIONS
 *
 ******************************************************************************/
typedef enum {
	PEDD = 0,
	PCIMAXNUM,
} synopsys_pci_t;


/**
 * Local RAM chunk id table
 */
typedef enum {
	CHUNK_ID_LLA,
	CHUNK_ID_LLB,
	CHUNK_ID_LLC,
	CHUNK_ID_LLD,
	CHUNK_ID_W0,
	CHUNK_ID_W1,
	CHUNK_ID_W2,
	CHUNK_ID_W3,
	CHUNK_ID_W4,
	CHUNK_ID_W5,
	CHUNK_ID_W6,
	CHUNK_ID_W7,
	CHUNK_ID_R0,
	CHUNK_ID_R1,
	CHUNK_ID_R2,
	CHUNK_ID_R3,
	CHUNK_ID_R4,
	CHUNK_ID_R5,
	CHUNK_ID_R6,
	CHUNK_ID_R7,
	CHUNK_ID_XDATA,
	CHUNK_ID_RSV0 = 0x1F, // RESERVED IDs 0x15 - 0x1F
	CHUNK_ID_DDR_W0,
	CHUNK_ID_DDR_R0
} CHUNK_ID;
/**
 * Supported commands of PEDD.
 */
typedef enum {
	DMA_CMD_INIT,       /**< DMA Initialization */
	DMA_CMD_STATE,      /**< Get channel state */
	DMA_CMD_DW,         /**< DMA write */
	DMA_CMD_DR,         /**< DMA read */
	DMA_CMD_DQF,        /**< Flush queuing DMA jobs */
	DMA_CMD_ABORT,      /**< Abort undergoing DMA transfer */
	DMA_CMD_STOP,       /**< Stop undergoing DMA transfer */
	DMA_CMD_RESUME,     /**< Resume stopped DMA transfer */
	DMA_CMD_MDUMP,      /**< Dump buffers */
	DMA_CMD_MADDR,      /**< Get addresses allocated for DMA device */
	DMA_CMD_RDUMP,      /**< Dump DMA registers address and data of interest */
	DMA_CMD_LLDUMP,     /**< Dump linked list elements */
	DMA_CMD_PERF,
	DMA_CMD_XPERF,	    /**< Get xData Performance data */
	DMA_CMD_XSTART,     /**< Start xData transmission */
	DMA_CMD_XSTOP,      /**< Stop xData transmission */
	DMA_CMD_XSTATUS,    /**< Get xData xfer status */
} DMA_CMD;

/**
 * Enumeration of data check status. Data check is performed after a transfer
 * is finished if requested.
 */
typedef enum {
	DATA_CHECK_NONE,    /**< No data check need to be performed */
	DATA_CHECK_UNKNOWN, /**< Transfer is running, so the result is unknown */
	DATA_CHECK_READY,   /**< Transfer is finished, ready to be checked. */
	DATA_CHECK_FAILURE, /**< Data check is failed */
	DATA_CHECK_SUCCESS  /**< Data check is successful */
} DMA_DATA_CHECK;

/**
* The end of a channel. Each channel has two ends, one for source and another
* for destination.
*/
typedef enum {
	DMA_CHAN_HST = 0,    /**< Host PC Buffer */
	DMA_CHAN_LOC,        /**< Local Buffer on FPGA */
	DMA_CHAN_DDR         /**< Local Buffer on DDR */
} DMA_CHAN_END;

/**
 * DMA channel information, as part of device extension, can be shared among
 * IRPs.
 * @note This data structure is used for application specific purpose. It has
 * nothing to do with the CHAN_INFO structure in HAL.
 */
typedef struct {
	UINT32 BlockSize;   	 /**< Block size */
	UINT64 TotalSize;   	 /**< Total transfer size of one request */
	DMA_DATA_CHECK DataCheck;/**< Data check status */
	UINT32 Pattern;     	 /**< Data pattern used by source data generation */
	UINT32 PatternIncre;	 /**< Increment of data pattern used by source data generation */
	UINT64 tick_count_start;
	UINT64 tick_count_stop;
} CHAN_INFO_EX, *PCHAN_INFO_EX;

/**
 * The PEDD command data structure. It contains the requested command and its
 * arguments.
 */
typedef struct {
	DMA_CMD Cmd;        /**< Command */
	UINT8 Chan;         /**< Channel number. Value can be 0 to 7. */
	DMA_DIRC Dirc;      /**< Write or read. */
	BOOLEAN Upstream;   /**< Upstream or downstream. Used by xData only .*/
	DMA_CHAN_END End;   /**< Source or destination */
	UINT8 LLElements;   /**< Numbers of linked list element, including both data
	                     element and link element */
	UINT8 WaterMark;    /**< Watermark element index */
	UINT8 Weight;       /**< The weight of the channel */
	UINT32 Len;         /**< Data length for one block transfer */
	UINT32 Recyc;       /**< How many times recycle the linked list */
	UINT32 Offset;      /**< Offset to start address of the memory */
	BOOLEAN Chk;        /**< Check data or not after the transfer */
	BOOLEAN Queue;      /**< Queue up the request or not */
	BOOLEAN Ddr;        /**< Using DDR memory or not */
	char ResMsg[MAX_MSG_SIZE];/**< Storage for holding the command response. */
} CMD_INFO, *PCMD_INFO;

/**
 *  Customized device extension
 */
typedef struct _DEVICE_EXTENSION {
	DMA_RSC DmaRsc;            /**< DMA resource object */
	UINT32 ChunkLen;           /**< Length of memory chunk for each channel */
	PHYADDR RamPa;             /**< The physical address of local RAM */
	PUINT32 RamVa;             /**< The virtual address of local RAM */
	UINT32 RamSize;            /**< The size of local RAM */
	PHYADDR IoMemPa[16];       /**< The physical address of buffer on system memory */
	PUINT32 IoMemVa[16];       /**< The virtual address of buffer on system memory */
	CHAN_INFO_EX WrChInfEx[8]; /**< The information for all write channels */
	CHAN_INFO_EX RdChInfEx[8]; /**< The information for all read channels */
	XDATA_RSC xDataRsc;        /**< Extra Data parameters */
	UINT32 ReadMTU;            /**< Max PCI packet payload size */
	UINT32 ReadMTUBits;
	UINT32 WriteMTU;
	UINT32 WriteMTUBits;
	dma_addr_t dma_bus[16];
} DEVICE_EXTENSION, *PDEVICE_EXTENSION;


/**
 *  pedd device structure
 */
struct pedd_dev {
	struct pci_dev	*pdev;
	struct cdev cdev;
	
	dev_t devno;
	
	void *mmio_addr;
	void *mmio_start;
	unsigned long mmio_len;
	int irq;
	void * __iomem bar[PEDD_BAR_NUM];
	PHYADDR RegSpacePa[PEDD_BAR_NUM];
	unsigned long bar_len[PEDD_BAR_NUM];
	
	int open_flag;
	int in_use;
	int got_regions;

	/* board info */
	uint8_t revision;
	uint8_t irq_pin;
	uint8_t irq_line;
	uint16_t sub_vendor_id;
	uint16_t sub_system_id;
	uint16_t vendor_id;
	uint16_t device_id;

	struct class *drv_cls;
	struct tasklet_struct tlet;

	DEVICE_EXTENSION pedd_ext;

	uint32_t irq_count;
};


/*******************************************************************************
 *
 * FUNCTION DECLARATION
 *
 ******************************************************************************/
static int  pedd_cdev_init(struct pedd_dev *peddev);
static void	pedd_priv_init(struct pedd_dev *priv);
static void	pedd_priv_exit(struct pedd_dev *priv);
static irqreturn_t pedd_isr(int irq, void *dev_id);

loff_t pedd_llseek(struct file *filp, loff_t off, int whence);
static ssize_t pedd_read(struct file *filp, char __user *buf, size_t count,
	                         loff_t *f_pos);
static ssize_t pedd_write(struct file *filp, const char __user *buf, size_t count,
                             loff_t *f_pos);
static ssize_t char_dma_read_write(struct file *filp, char __user *buf,
		size_t count, loff_t *pos, int dir_to_dev);
int akida_bar_map_iATU(struct pci_dev *pdev);


int akida_regs_mmap(struct file *filp, struct vm_area_struct *vma);
static long    pedd_ioctl(struct file *filp, unsigned int cmd, unsigned long arg);
static int     pedd_open(struct inode *inode, struct file *filp);
static int     pedd_release(struct inode *inode, struct file *filp);
static int  pedd_probe(struct pci_dev *pdev, const struct pci_device_id *pci_id);
static void pedd_remove(struct pci_dev *pci_dev);
void pedd_Read_Header(struct pci_dev *pdev, PCI_COMMON_CONFIG *data);
void pedd_rw_pci_cfg(struct pci_dev *pdev, INT d, UINT32 length, UINT32 offset, UCHAR *val);
void pedd_rw_pci_bar(INT d, UINT32 length, void *bar_addr, UINT32 offset, UCHAR *val);
int pedd_rw_pci_hdr(struct pci_dev *pdev, INT rc, INT d, UINT32 length, UINT32 offset, UCHAR *val);
static BOOLEAN ParseCommand(CHAR *msg, PCMD_INFO cmdinfo);
static UINT32 IoctlCommand(struct pedd_dev *priv, PDEVICE_EXTENSION pDevExt, PCMD_INFO pcmd);
static void DumpLocRam( const PUINT32 addr, UINT32 len, PUINT32 res );
static void DumpSysMem( const PUINT32 addr, UINT32 len, PUINT32 res );
static PHYADDR GetBaseDataPAddr(PDEVICE_EXTENSION pDevExt, DMA_DIRC dirc,UINT8 chan, DMA_CHAN_END end);
static PUINT32 GetBaseDataVAddr(PDEVICE_EXTENSION pDevExt, DMA_DIRC dirc, UINT8 chan, DMA_CHAN_END end);
static BOOLEAN CheckData(PDEVICE_EXTENSION pDevExt, DMA_DIRC dirc, UINT8 chan); 
void DataCheckDpc(unsigned long arg);
static UINT32 AllocProtocolMem(struct pedd_dev *priv, UINT32 memSize, PUINT32 *pvaddr, dma_addr_t *dma_bus, PHYADDR *ppaddr);
static VOID FreeProtocolMem(struct pedd_dev *priv, UINT32 memSize, PUINT32 *vaddr, dma_addr_t *dma_bus);
static UINT32 BatchAllocDmaMem(struct pedd_dev *priv, int wrChans, int rdChans);
static void PeddXferDoneCallback(DMA_DIRC dirc, UINT8 chan, PVOID ctx);
static UINT32 IoctlDmaInit(struct pedd_dev *priv, PDEVICE_EXTENSION pDevExt, PCMD_INFO pcmd);
static UINT32 IoctlDmaState(PDEVICE_EXTENSION pDevExt, PCMD_INFO pcmd);
static UINT32 IoctlDmaRW(PDEVICE_EXTENSION pDevExt, PCMD_INFO pcmd);
static UINT32 IoctlDmaAbort(PDEVICE_EXTENSION pDevExt, PCMD_INFO pcmd);
static UINT32 IoctlDmaResume(PDEVICE_EXTENSION pDevExt, PCMD_INFO pcmd);
static UINT32 IoctlDmaStop(PDEVICE_EXTENSION pDevExt, PCMD_INFO pcmd);
static UINT32 IoctlMemDump(PDEVICE_EXTENSION pDevExt, PCMD_INFO pcmd);
static UINT32 IoctlPerf(PDEVICE_EXTENSION pDevExt, PCMD_INFO pcmd);
static UINT32 IoctlDmaLLDump(PDEVICE_EXTENSION pDevExt, PCMD_INFO pcmd);
static UINT32 IoctlDmaRegsDump(PDEVICE_EXTENSION pDevExt, PCMD_INFO pcmd);
static UINT32 IoctlDmaFlushQueue(PDEVICE_EXTENSION pDevExt, PCMD_INFO pcmd);
static UINT32 IoctlxDataStatus(PDEVICE_EXTENSION pDevExt, PCMD_INFO pcmd);
static UINT32 IoctlxDataStopXfer(PDEVICE_EXTENSION pDevExt, PCMD_INFO pcmd);
static UINT32 IoctlxDataStartXfer(PDEVICE_EXTENSION pDevExt, PCMD_INFO pcmd);
static UINT32 IoctlBufAddr(PDEVICE_EXTENSION pDevExt, PCMD_INFO pcmd);
static void FillLocRam(PUINT32 addr, UINT32 len, UINT32 seed, UINT32 incr);
static void FillSysMem(PUINT32 addr, UINT32 len, UINT32 seed, UINT32 incr);
static void GenSrcData(PDEVICE_EXTENSION pDevExt, DMA_DIRC dirc, UINT8 chan,UINT32 pat, UINT32 patInc, BOOLEAN isDdr);
static BOOLEAN PeddProducerFunc(PHYADDR *src, PHYADDR *dst,UINT32 *size, DMA_DIRC dirc, UINT8 chan, PVOID context);


#endif  //__PEDD_H__
