// SPDX-License-Identifier: GPL-2.0
/*
 * Copyright (c) 2018 Synopsys, Inc. and/or its affiliates.
 * Synopsys DesignWare eDMA controller driver.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/string.h>
#include <linux/slab.h>
#include <linux/fs.h>
#include <linux/types.h>
#include <linux/cdev.h>
#include <linux/pci.h>
#include <linux/device.h>
#include <linux/jiffies.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/uaccess.h>
#include <linux/msi.h>
#include <uapi/linux/pci_regs.h>
#include <asm/uaccess.h>
#include "pedd_dfx.h"
#include "pedd_cmd.h"
#include "pedd_hal_dma.h"
#include "pedd_dma.h"
#include "pedd_perf.h"
#include "pedd_xdata.h"
#include "pedd_types.h"
#include "ped_ioctl.h"
#include "pedd.h"
#include "akida-sgm.h"

MODULE_LICENSE("GPL v2");
MODULE_AUTHOR("Synopsys");
MODULE_DESCRIPTION("Synopsys DesignWare eDMA PCIe driver");
MODULE_VERSION("1.0");

// xData feature
static short int xData = XDATA;
module_param(xData, short, 0644);
MODULE_PARM_DESC(xData, "xData enable\n");

// eDMA feature
static short int eDMA = EDMA;
module_param(eDMA, short, 0644);
MODULE_PARM_DESC(eDMA, "eDMA enable\n");

static bool MSI_only = MSI_ONLY;
module_param(MSI_only, bool, 0644);
MODULE_PARM_DESC(MSI_only, "MSI-X disable\n");

// xData location
static short int xData_bar = XDATA_BAR;
module_param(xData_bar, short, 0644);
MODULE_PARM_DESC(xData_bar, "xData BAR location\n");

// eDMA memory location
static short int eDMA_mem_bar = EDMA_MEM_BAR;
module_param(eDMA_mem_bar, short, 0644);
MODULE_PARM_DESC(eDMA_mem_bar, "eDMA memory BAR location\n");

// eDMA register location
static short int eDMA_regs_bar = EDMA_REGS_BAR;
module_param(eDMA_regs_bar, short, 0644);
MODULE_PARM_DESC(eDMA_regs_bar, "eDMA registers BAR location\n");

static long eDMA_regs_off = EDMA_REGS_OFF;
module_param(eDMA_regs_off, long, 0644);
MODULE_PARM_DESC(eDMA_regs_off, "eDMA registers offset location (in bytes)\n");

static bool HW_performance_workaround = HW_PERFORMANCE_WORKAROUND;
module_param(HW_performance_workaround, bool, 0644);
MODULE_PARM_DESC(HW_performance_workaround, "Enables HW performance workaround\n");

static const struct pci_device_id pedd_pci_tbl[] = {
	{ PCI_DEVICE(PEDD_TARGET_VENDOR_ID, PEDD_TARGET_DEVICE_ID)},
	{ }
};
MODULE_DEVICE_TABLE (pci, pedd_pci_tbl);

static struct file_operations pedd_fops = {
	.owner          = THIS_MODULE,
	.llseek			= pedd_llseek,
	.read           = pedd_read,
	.write          = pedd_write,
	.mmap		= akida_regs_mmap,
	.unlocked_ioctl = pedd_ioctl,
	.open           = pedd_open,
	.release        = pedd_release,
};

static struct pci_driver pedd_pci_driver = {
	.name     = DRVER_NAME,
	.id_table = pedd_pci_tbl,
	.probe    = pedd_probe,
	.remove   = pedd_remove,
};

#define XDATA_OFFSET 0x06000000
struct pedd_dev *pedd_devices = NULL;
dev_t pedd_devno;
int pedd_cdev_major = PEDD_CDEV_MAJOR;
int pedd_cdev_minor = 0;
int pedd_nr_devs    = PEDD_NR_DEVS;


/**
 * Implements HAL callback function PRODUCER_FUNC().
 *
 * This function is responsible for generating new data for linked list element.
 * The source and destination are wrapped around within the chunk of memory
 * owned by specified channel.
 *
 * @param[in,out] src    The source of the transfer.
 * @param[in,out] dst    The destination of the transfer.
 * @param[in,out] size   The size of the data need to be transferred.
 * @param[in] dirc       The direction of the channel.
 * @param[in] chan       The number of the channel.
 * @param[in] context    The callback context.
 * @return True if new data is generated successfully; False otherwise.
 * @see PRODUCER_FUNC
 * @ingroup hal_caller
 */
static BOOLEAN PeddProducerFunc(PHYADDR *src, PHYADDR *dst,
				UINT32 *size, DMA_DIRC dirc,
				UINT8 chan, PVOID context)
{
	BOOLEAN isDdr;
	PHYADDR bsrc;
	PHYADDR bdst;
	PDEVICE_EXTENSION pDevExt = (PDEVICE_EXTENSION)context;
	PCHAN_INFO_EX pChInfEx = (dirc == DMA_WRITE) ? &pDevExt->WrChInfEx[chan]
						     : &pDevExt->RdChInfEx[chan];

	/* We transfer same amount of data each time */
	*size = pChInfEx->BlockSize;
	isDdr = pChInfEx->PatternIncre > 2;

	bsrc = GetBaseDataPAddr(pDevExt, dirc, chan, GET_CHAN_SRC(dirc, isDdr));
	bdst = GetBaseDataPAddr(pDevExt, dirc, chan, GET_CHAN_DST(dirc, isDdr));

	if (src->LowPart == 0 && src->HighPart == 0) {
		*src = bsrc;
		*dst = bdst;
	} else {
		UINT32 boundry = bsrc.LowPart + pDevExt->ChunkLen * 4;
		/* If the address reachs the boundry, wrap it back. The assumption here
		 * is that source and destination chunk of memory have the same size.
		 */
		src->LowPart += *size;
		dst->LowPart += *size;
		if (src->LowPart > boundry - (*size)) {
			*src = bsrc;
			*dst = bdst;
		}
	}

	return TRUE;
}


/**
 * Fill the local RAM with data derived from given pattern and increment,
 * using regiser access method.
 *
 * @param[in] addr  The address of local RAM
 * @param[in] len   The length of data
 * @param[in] seed  The seed data pattern
 * @param[in] incr  The increment value
 */
static void FillLocRam(PUINT32 addr, UINT32 len, UINT32 seed, UINT32 incr)
{
	UINT32 data = seed;
	UINT32 i;

	for (i = 0; i < len; i++) {
#if 1
		/* This is a special setting for current prototype, since there are
		 * only 2K physical memory available for each channel, no matter how
		 * many addresses are occupied.
		 */
		if (i == 512)
		    break;
#endif
		PEDD_WRITE_UINT32(addr, data);
		data += incr;
		addr++;
	}
}


/**
 * Fill the system memory with data derived from given pattern and increment.
 *
 * Similiar to FillLocRam but use direct memory access.
 *
 * @param[in] addr  The address of system memory
 * @param[in] len   The length of data
 * @param[in] seed  The seed data pattern
 * @param[in] incr  The increment value
 */
static void FillSysMem(PUINT32 addr, UINT32 len, UINT32 seed, UINT32 incr)
{
	UINT32 data = seed;
	UINT32 i;

	for (i = 0; i < len; i++) {
#if 1
	/* This is a special setting for current prototype, since there are
	 * only 2K physical memory available for each channel, no matter how
	 * many addresses are occupied.
	 */
		if (i % 512 == 0)
			data = seed;
#endif
		*addr = data;
		data += incr;
		addr++;
	}
}

/**
 * Generate source data for given channel.
 *
 * This function need to be called before DMA transfer is started so that we
 * can possibly check the destination data against data pattern.
 *
 * @param[in] pDevExt   The device extention object.
 * @param[in] dirc      The direction of the channel.
 * @param[in] chan      The number of the channel.
 * @param[in] pat       The data pattern.
 * @param[in] patInc    The increment value.
 * @param[in] isDdr     Whether use DDR memory.
 */
static void GenSrcData(PDEVICE_EXTENSION pDevExt, DMA_DIRC dirc, UINT8 chan,
		       UINT32 pat, UINT32 patInc, BOOLEAN isDdr)
{
	PUINT32 addr = GetBaseDataVAddr(pDevExt, dirc, chan,
					GET_CHAN_SRC(dirc, isDdr));
	UINT32 len = pDevExt->ChunkLen;

	if (dirc == DMA_WRITE)
		FillLocRam(addr, len, pat, patInc);
	else
		FillSysMem(addr, len, pat, patInc);
}


/**
 * Dump local RAM content, using register access method.
 *
 * @param[in] addr  The address of the local RAM
 * @param[in] len   The length of data
 * @param[out] res  The pointer to an array to hold the returned data. Please
 *                  be sure the array has enough space to hold the data.
 */
static void DumpLocRam(const PUINT32 addr, UINT32 len, PUINT32 res)
{
	while (len) {
		len--;
		res[len] = PEDD_READ_UINT32(addr + len);
	}
}

/**
 * Dump system memory content.
 *
 * Similiar to DumpLocRam, but use direct memory access.
 *
 * @param[in] addr  The address of local RAM
 * @param[in] len   The length of data
 * @param[out] res  The pointer to an array to hold the returned data. Please
 *                  be sure the array has enough space to hold the data.
 */
static void DumpSysMem(const PUINT32 addr, UINT32 len, PUINT32 res)
{
	while (len) {
		len--;
		res[len] = addr[len];
	}
}

/**
 * This function is similiar to GetBaseDataVAddr, but it returns the physical
 * address instead of virtual address.
 *
 * @param[in] pDevExt   Device extension object
 * @param[in] dirc      The direction of the channel.
 * @param[in] chan      The number of the channel.
 * @param[in] end       The end of the channel.
 * @return the physical address of starting location for given channel.
 * @see GetBaseDataVAddr
 */
static PHYADDR GetBaseDataPAddr(PDEVICE_EXTENSION pDevExt, DMA_DIRC dirc,
	UINT8 chan, DMA_CHAN_END end)
{
	PHYADDR addr = {{0} };

	if ((dirc == DMA_WRITE && chan >= pDevExt->DmaRsc.WrChans) || (dirc
	    == DMA_READ && chan >= pDevExt->DmaRsc.RdChans))
		return addr;

	if (dirc == DMA_READ)
		chan += 8;

	if (end == DMA_CHAN_HST) {
		addr = pDevExt->IoMemPa[chan];
	} else {
		int baseId = end == DMA_CHAN_DDR ? CHUNK_ID_DDR_W0
						 : CHUNK_ID_W0;
		addr = pDevExt->RamPa;
		addr.LowPart += pDevExt->ChunkLen * (baseId + chan) * 4;
	}

	return addr;
}

/**
 * Get the base data virtual address of given channel.
 *
 * Both the system memory and local memory are segmented to chunks for all
 * channels, so each channel has its unique block of memory as source and
 * unique block of memory as destination. This function returns the starting
 * address of that block of memory for given channel.
 *
 * @param[in] pDevExt   Device extension object
 * @param[in] dirc      The direction of the channel.
 * @param[in] chan      The number of the channel.
 * @param[in] end       The end of the channel.
 * @return the virtual address of starting location for given channel.
 */
static PUINT32 GetBaseDataVAddr(PDEVICE_EXTENSION pDevExt, DMA_DIRC dirc,
				UINT8 chan, DMA_CHAN_END end)
{
	PUINT32 addr = NULL;

	if ((dirc == DMA_WRITE && chan >= pDevExt->DmaRsc.WrChans) || (dirc
	    == DMA_READ && chan >= pDevExt->DmaRsc.RdChans))
		return addr;

	if (dirc == DMA_READ)
		chan += 8;

	if (end == DMA_CHAN_HST) {
		addr = pDevExt->IoMemVa[chan];
	} else {
		int baseId = end == DMA_CHAN_DDR ? CHUNK_ID_DDR_W0 : CHUNK_ID_W0;
		addr = pDevExt->RamVa + pDevExt->ChunkLen * (baseId + chan);
	}

    return addr;
}

/**
 * Check data transferred by DMA.
 *
 * This function should only be called after the transfer is finished. Also,
 * since this function takes time to compare the destination data and the
 * expected pattern, it should not be called directly from ISR. Instead,
 * calling it in DPC is appropriate.
 *
 * @param[in] pDevExt   The device extension object.
 * @param[in] dirc      The direction of the channel.
 * @param[in] chan      The number of the channel.
 * @return True if the check passed; False otherwise.
 * @see HalDmaState
 * @ingroup hal_caller
 */
static BOOLEAN CheckData(PDEVICE_EXTENSION pDevExt, DMA_DIRC dirc, UINT8 chan)
{
	UINT64 xferred;
	PCHAN_INFO_EX pChInfEx;
	UINT32 len;
	BOOLEAN isDdr;
	PUINT32 srcAddr;
	PUINT32 dstAddr;
	UINT32 srcData, dstData;
	BOOLEAN ret = TRUE;
	PDMA_RSC pdrsc = &pDevExt->DmaRsc;

	/* Firstly check the state of given channel to see how many data has been
	* transferred. Data check only perform checks on that amount of data.
	*/
	HalDmaState(pdrsc, dirc, chan, NULL, NULL, &xferred, NULL, NULL);

	pChInfEx = (dirc == DMA_WRITE) ? &pDevExt->WrChInfEx[chan]
	    : &pDevExt->RdChInfEx[chan];
	len = ((PEDD_DIV64(xferred, 4)) > pDevExt->ChunkLen)
	    ? pDevExt->ChunkLen - (pDevExt->ChunkLen % (pChInfEx->BlockSize/4))
	    : (UINT32)xferred / 4;
	isDdr = pChInfEx->PatternIncre > 2;
	srcAddr = GetBaseDataVAddr(pDevExt, dirc, chan, GET_CHAN_SRC(dirc, isDdr));
	dstAddr = GetBaseDataVAddr(pDevExt, dirc, chan,
	    GET_CHAN_DST(dirc, isDdr));

	DTRACE("RAM data pattern check: len=%d\n", len);

	pChInfEx->DataCheck = DATA_CHECK_SUCCESS;
	while (len) {
		if (dirc == DMA_READ) {
			DumpLocRam(dstAddr, 1, &dstData);
			DumpSysMem(srcAddr, 1, &srcData);
		} else {
			DumpSysMem(dstAddr, 1, &dstData);
			DumpLocRam(srcAddr, 1, &srcData);
		}
		if (srcData != dstData) {
			DTRACE("len=%d [0x%x] %08x != %08x(pat)\n", len,
			    (UINT32) (dstAddr -
			    GetBaseDataVAddr(pDevExt, dirc, chan,
			    GET_CHAN_DST(dirc, isDdr))),
			    dstData, srcData);
			pChInfEx->DataCheck = DATA_CHECK_FAILURE;
			ret = FALSE;
			break;
		}
		srcAddr++;
		dstAddr++;
		len--;
	}

	return ret;
}


void DataCheckDpc(unsigned long arg)
{
	struct pedd_dev *priv = (struct pedd_dev *)arg;
	PDEVICE_EXTENSION pDevExt = &priv->pedd_ext;
	PCHAN_INFO_EX pChInfEx;
	DMA_DIRC dirc;
	UINT8 chan;

	/* Traverse all channels to see if need to do data check */
	dirc = DMA_WRITE;
	for (chan = 0; chan < pDevExt->DmaRsc.WrChans; chan++) {
		pChInfEx = &pDevExt->WrChInfEx[chan];
		if (pChInfEx->DataCheck == DATA_CHECK_READY)
			CheckData(pDevExt, dirc, chan);
	}
	dirc = DMA_READ;
	for (chan = 0; chan < pDevExt->DmaRsc.RdChans; chan++) {
		pChInfEx = &pDevExt->RdChInfEx[chan];
		if (pChInfEx->DataCheck == DATA_CHECK_READY)
			CheckData(pDevExt, dirc, chan);
	}
}

static UINT32 AllocProtocolMem(struct pedd_dev *priv, UINT32 memSize, PUINT32 *pvaddr, dma_addr_t *dma_bus, PHYADDR *ppaddr)
{
	UINT32 status = STATUS_SUCCESS;
	UINT32 i;
	void *dma_buf;
	PUCHAR memVa;

	INFO_PRINT("Allocating memory for I/O buffers\n");

	dma_buf = pci_alloc_consistent(priv->pdev, memSize, dma_bus);

	if (HW_performance_workaround)
		*dma_bus = (dma_addr_t)(priv->bar[EDMA_MEM_BAR] + XDATA_OFFSET);

	if (!dma_buf) {
		ERROR_PRINT("error pci_alloc_consistent\n");
		status = STATUS_FAIL;
		*pvaddr = 0;
		ppaddr->LowPart = 0;
		ppaddr->HighPart = 0;
		goto CLEANUP;
	}

	*pvaddr = (PUINT32)dma_buf;
	ppaddr->LowPart = cpu_to_le32(pci_dma_l(*dma_bus));
	ppaddr->HighPart = cpu_to_le32(pci_dma_h(*dma_bus));

	/* Initialize the allocated memory */
	memVa = (PUCHAR)dma_buf;
	for (i = 0; i < memSize; i++)
		*memVa++ = 0xcd;

CLEANUP:

	return status;
}

static VOID FreeProtocolMem(struct pedd_dev *priv, UINT32 memSize, PUINT32 *vaddr, dma_addr_t *dma_bus)
{
	if (*vaddr == NULL)
		return;

	pci_free_consistent(priv->pdev, memSize, *vaddr, *dma_bus);

}

static UINT32 BatchAllocDmaMem(struct pedd_dev *priv, int wrChans, int rdChans)
{
	UINT32 status = STATUS_SUCCESS;
	int i = 0;

	for (i = 0; status == STATUS_SUCCESS && i < wrChans; i++) {
		status = AllocProtocolMem(priv, priv->pedd_ext.ChunkLen * 4,
					  &priv->pedd_ext.IoMemVa[i],
					  &priv->pedd_ext.dma_bus[i],
					  &priv->pedd_ext.IoMemPa[i]);
	}

	for (i = 8; status == STATUS_SUCCESS && i < 8 + rdChans; i++) 	{
		status = AllocProtocolMem(priv, priv->pedd_ext.ChunkLen * 4,
					  &priv->pedd_ext.IoMemVa[i],
					  &priv->pedd_ext.dma_bus[i],
					  &priv->pedd_ext.IoMemPa[i]);
	}

	/* Free all previously allocated memory if any allocation is failed. */
	if (status != STATUS_SUCCESS) {
		while (--i >= 0)
			FreeProtocolMem(priv, priv->pedd_ext.ChunkLen * 4,
				&priv->pedd_ext.IoMemVa[i], &priv->pedd_ext.dma_bus[i]);
	}

	return status;
}


static void PeddXferDoneCallback(DMA_DIRC dirc, UINT8 chan, PVOID ctx)
{
	PDEVICE_EXTENSION pDevExt = (PDEVICE_EXTENSION)ctx;
	struct pedd_dev *priv = container_of(pDevExt, struct pedd_dev, pedd_ext);

	PCHAN_INFO_EX pChInfEx = dirc == DMA_WRITE ? &pDevExt->WrChInfEx[chan]
	    : &pDevExt->RdChInfEx[chan];

	if (dirc == DMA_WRITE)
 	   DTRACE("PeddXferDoneCallback - write ISR ack\n");
	else
 	   DTRACE("PeddXferDoneCallback - read ISR ack\n");
	pChInfEx->tick_count_stop = jiffies;

	/* if one channel is finished and data check need to be performed, kick off
	* DPC to do that. */
	if (pChInfEx->DataCheck == DATA_CHECK_UNKNOWN) {
		pChInfEx->DataCheck = DATA_CHECK_READY;

		tasklet_schedule(&priv->tlet);
		DTRACE("Transfer on %s channel %d is finished.",
		       dirc == DMA_WRITE ? "write" : "read", chan);
	}
	DTRACE("PeddXferDoneCallback: Done\n");
}


/**
 * Handler for DMA_CMD_MADDR IOCTL command.
 *
 * This function doesn't call any HAL API. It print out the size, virtual and
 * physical address of application related memory.
 *
 * @param[in] pDevExt   The device extension object.
 * @param[in,out] pcmd  The command object containing all arguments of requested
 *                      command. Returning messages are also stored in this
 *                      object.
 * @return NTSTATUS value to indicate success or failure.
 */
static UINT32 IoctlBufAddr(PDEVICE_EXTENSION pDevExt, PCMD_INFO pcmd)
{
	UINT32 status = STATUS_SUCCESS;
	struct pedd_dev *priv = container_of(pDevExt, struct pedd_dev, pedd_ext);

	DTRACE("ENTER\n");

	DPRINTF(pcmd->ResMsg, MAX_MSG_SIZE,
		"BAR0: size=%ld KB va=0x%p\n"
		"BAR2: size=%ld KB va=0x%p\n"
		"\nBuffers for each channel (%d KB per buffer):\n"
		"   WR_loc     WR_sys     RD_loc    RD_sys\n"
		"------------------------------------------------\n"
		"0: 0x%08X 0x%08X 0x%08X 0x%08X\n"
		"1: 0x%08X 0x%08X 0x%08X 0x%08X\n"
		"2: 0x%08X 0x%08X 0x%08X 0x%08X\n"
		"3: 0x%08X 0x%08X 0x%08X 0x%08X\n"
		"4: 0x%08X 0x%08X 0x%08X 0x%08X\n"
		"5: 0x%08X 0x%08X 0x%08X 0x%08X\n"
		"6: 0x%08X 0x%08X 0x%08X 0x%08X\n"
		"7: 0x%08X 0x%08X 0x%08X 0x%08X\n",

		//Now use bar 0 and bar 2
		priv->bar_len[0] / 1024, priv->bar[0],
		priv->bar_len[2] / 1024, priv->bar[2],

		pDevExt->ChunkLen / 256,
		GetBaseDataPAddr(pDevExt, DMA_WRITE, 0, DMA_CHAN_LOC).LowPart,
		GetBaseDataPAddr(pDevExt, DMA_WRITE, 0, DMA_CHAN_HST).LowPart,
		GetBaseDataPAddr(pDevExt, DMA_READ, 0, DMA_CHAN_LOC).LowPart,
		GetBaseDataPAddr(pDevExt, DMA_READ, 0, DMA_CHAN_HST).LowPart,
		GetBaseDataPAddr(pDevExt, DMA_WRITE, 1, DMA_CHAN_LOC).LowPart,
		GetBaseDataPAddr(pDevExt, DMA_WRITE, 1, DMA_CHAN_HST).LowPart,
		GetBaseDataPAddr(pDevExt, DMA_READ, 1, DMA_CHAN_LOC).LowPart,
		GetBaseDataPAddr(pDevExt, DMA_READ, 1, DMA_CHAN_HST).LowPart,
		GetBaseDataPAddr(pDevExt, DMA_WRITE, 2, DMA_CHAN_LOC).LowPart,
		GetBaseDataPAddr(pDevExt, DMA_WRITE, 2, DMA_CHAN_HST).LowPart,
		GetBaseDataPAddr(pDevExt, DMA_READ, 2, DMA_CHAN_LOC).LowPart,
		GetBaseDataPAddr(pDevExt, DMA_READ, 2, DMA_CHAN_HST).LowPart,
		GetBaseDataPAddr(pDevExt, DMA_WRITE, 3, DMA_CHAN_LOC).LowPart,
		GetBaseDataPAddr(pDevExt, DMA_WRITE, 3, DMA_CHAN_HST).LowPart,
		GetBaseDataPAddr(pDevExt, DMA_READ, 3, DMA_CHAN_LOC).LowPart,
		GetBaseDataPAddr(pDevExt, DMA_READ, 3, DMA_CHAN_HST).LowPart,
		GetBaseDataPAddr(pDevExt, DMA_WRITE, 4, DMA_CHAN_LOC).LowPart,
		GetBaseDataPAddr(pDevExt, DMA_WRITE, 4, DMA_CHAN_HST).LowPart,
		GetBaseDataPAddr(pDevExt, DMA_READ, 4, DMA_CHAN_LOC).LowPart,
		GetBaseDataPAddr(pDevExt, DMA_READ, 4, DMA_CHAN_HST).LowPart,
		GetBaseDataPAddr(pDevExt, DMA_WRITE, 5, DMA_CHAN_LOC).LowPart,
		GetBaseDataPAddr(pDevExt, DMA_WRITE, 5, DMA_CHAN_HST).LowPart,
		GetBaseDataPAddr(pDevExt, DMA_READ, 5, DMA_CHAN_LOC).LowPart,
		GetBaseDataPAddr(pDevExt, DMA_READ, 5, DMA_CHAN_HST).LowPart,
		GetBaseDataPAddr(pDevExt, DMA_WRITE, 6, DMA_CHAN_LOC).LowPart,
		GetBaseDataPAddr(pDevExt, DMA_WRITE, 6, DMA_CHAN_HST).LowPart,
		GetBaseDataPAddr(pDevExt, DMA_READ, 6, DMA_CHAN_LOC).LowPart,
		GetBaseDataPAddr(pDevExt, DMA_READ, 6, DMA_CHAN_HST).LowPart,
		GetBaseDataPAddr(pDevExt, DMA_WRITE, 7, DMA_CHAN_LOC).LowPart,
		GetBaseDataPAddr(pDevExt, DMA_WRITE, 7, DMA_CHAN_HST).LowPart,
		GetBaseDataPAddr(pDevExt, DMA_READ, 7, DMA_CHAN_LOC).LowPart,
		GetBaseDataPAddr(pDevExt, DMA_READ, 7, DMA_CHAN_HST).LowPart
	);

	return status;
}


/**
 * Handler of DMA_CMD_XSTART IOCTL command.
 *
 * This function starts xData transfer
 *
 * @param[in] pDevExt   The device extension object.
 * @param[in,out] pcmd  The command object containing all arguments of requested
 *                      command. Returning messages are also stored in this
 *                      object.
 * @return NTSTATUS value to indicate success or failure.
 */
static UINT32 IoctlxDataStartXfer(PDEVICE_EXTENSION pDevExt, PCMD_INFO pcmd)
{
	UINT32  status = STATUS_SUCCESS;
	XDATA_RET xret;
	BOOLEAN   isWrite = (pcmd->Dirc == DMA_WRITE);

	DTRACE("ENTER\n");

	xret = xDataStartXfer(&pDevExt->xDataRsc, isWrite, pcmd->Upstream, pcmd->ResMsg);

	if (xret != XDATA_SUCCESS)
		status = STATUS_UNSUCCESSFUL;

	return status;
}


/**
 * Handler of DMA_CMD_XSTOP IOCTL command.
 *
 * This function stops xData transfer
 *
 * @param[in] pDevExt   The device extension object.
 * @param[in,out] pcmd  The command object containing all arguments of requested
 *                      command. Returning messages are also stored in this
 *                      object.
 * @return NTSTATUS value to indicate success or failure.
 */
static UINT32 IoctlxDataStopXfer(PDEVICE_EXTENSION pDevExt, PCMD_INFO pcmd)
{
	UINT32 status = STATUS_SUCCESS;
	XDATA_RET xret;

	DTRACE("ENTER\n");

	xret = xDataStopXfer(&pDevExt->xDataRsc, pcmd->Upstream, pcmd->ResMsg);

	if (xret != XDATA_SUCCESS)
		status = STATUS_UNSUCCESSFUL;

	return status;
}



/**
 * Handler of DMA_CMD_XSTATUS IOCTL command.
 *
 * This function stops xData transfer
 *
 * @param[in] pDevExt   The device extension object.
 * @param[in,out] pcmd  The command object containing all arguments of requested
 *                      command. Returning messages are also stored in this
 *                      object.
 * @return NTSTATUS value to indicate success or failure.
 */
static UINT32 IoctlxDataStatus(PDEVICE_EXTENSION pDevExt, PCMD_INFO pcmd)
{
	UINT32 status = STATUS_SUCCESS;
	XDATA_RET xret;

	DTRACE("ENTER\n");

	xret = xDataStatus(&pDevExt->xDataRsc, pcmd->ResMsg);

	if (xret != XDATA_SUCCESS)
		status = STATUS_UNSUCCESSFUL;

	return status;
}


/**
 * Handler of DMA_CMD_DQF IOCTL command.
 *
 * This function calls HAL flush queue API to kick off any queuing jobs. If
 * there is no queuing jobs, this function takes no effect.
 *
 * @param[in] pDevExt   The device extension object.
 * @param[in,out] pcmd  The command object containing all arguments of requested
 *                      command. Returning messages are also stored in this
 *                      object.
 * @return NTSTATUS value to indicate success or failure.
 * @see HalDmaFlushQueue
 * @ingroup hal_caller
 */
static UINT32 IoctlDmaFlushQueue(PDEVICE_EXTENSION pDevExt, PCMD_INFO pcmd)
{
	UINT32 status = STATUS_SUCCESS;
	HAL_DMA_RET ret;

	DTRACE("ENTER\n");

	/* Call HAL flush queue API to kick off any queuing transfer. */
	ret = HalDmaFlushQueue(&pDevExt->DmaRsc);

	/* See if the request is successful. */
	if (ret == DMA_SUCCESS) {
		DPRINTF(pcmd->ResMsg, MAX_MSG_SIZE,
			"OK! All queuing jobs are kicked off.\n");
	} else {
		DEPRINTF(pcmd->ResMsg, MAX_MSG_SIZE,
			"Failed to flush queuing DMA jobs:\n%s\n",
		HalDmaGetErrInfo(ret));
		status = STATUS_UNSUCCESSFUL;
	}

	return status;
}

/**
 * Handler of DMA_CMD_RDUMP IOCTL command.
 *
 * This function calls HAL register dump API to read out the addresses and
 * contents of registers of interest.
 *
 * @param[in] pDevExt   The device extension object.
 * @param[in,out] pcmd  The command object containing all arguments of requested
 *                      command. Returning messages are also stored in this
 *                      object.
 * @return NTSTATUS value to indicate success or failure.
 * @see HalDmaDumpReg
 * @ingroup hal_caller
 */
static UINT32 IoctlDmaRegsDump(PDEVICE_EXTENSION pDevExt, PCMD_INFO pcmd)
{
	UINT32 status = STATUS_SUCCESS;
	HAL_DMA_RET ret;

	DTRACE("ENTER\n");

	/* Call HAL register dump API to dump registers of interest. */
	ret = HalDmaDumpReg(&pDevExt->DmaRsc, pcmd->Dirc, pcmd->Chan,
	    pcmd->ResMsg);

	/* See if the request is successful. */
	if (ret != DMA_SUCCESS) {
		DEPRINTF(pcmd->ResMsg, MAX_MSG_SIZE,
			 "Failed to dump DMA registers:\n%s\n",
			 HalDmaGetErrInfo(ret));
		status = STATUS_UNSUCCESSFUL;
	}

	return status;
}

/**
 * Handler of DMA_CMD_LLDUMP IOCTL command.
 *
 * This function calls HAL linked list dump API to read out the addresses and
 * contents of linked list element.
 *
 * @param[in] pDevExt   The device extension object.
 * @param[in,out] pcmd  The command object containing all arguments of requested
 *                      command. Returning messages are also stored in this
 *                      object.
 * @return NTSTATUS value to indicate success or failure.
 * @see HalDmaDumpLL
 * @ingroup hal_caller
 */
static UINT32 IoctlDmaLLDump(PDEVICE_EXTENSION pDevExt, PCMD_INFO pcmd)
{
	UINT32 status = STATUS_SUCCESS;
	HAL_DMA_RET ret;

	DTRACE("ENTER\n");

	/* Call HAL LL dump API to dump linked list of given channel. */
	ret = HalDmaDumpLL(&pDevExt->DmaRsc, pcmd->Dirc, pcmd->Chan,
	    pcmd->ResMsg);

	/* See if the request is successful. */
	if (ret != DMA_SUCCESS) {
		DEPRINTF(pcmd->ResMsg, MAX_MSG_SIZE,
			"Failed to dump Linked List:\n%s\n",
			HalDmaGetErrInfo(ret));
		status = STATUS_UNSUCCESSFUL;
	}

	return status;
}

/**
 * Handler of DMA_CMD_PERF IOCTL command.
 *
 * This function dumps performance registers. These values can be used to
 * measure DMA, xData, and total transfer performance.
 *
 * @param[in] pDevExt   The device extension object.
 * @param[in,out] pcmd  The command object containing all arguments of requested
 *                      command. Returning messages are also stored in this
 *                      object.
 * @return NTSTATUS value to indicate success or failure.
 */
static UINT32 IoctlPerf(PDEVICE_EXTENSION pDevExt, PCMD_INFO pcmd)
{

	UINT32 status = STATUS_SUCCESS;

	PPERF_REGS pperfRegs = (PPERF_REGS)pDevExt->DmaRsc.DmaRegs;
	UCHAR tick;
	UINT64 time, rxTlp, rxTlpNum, rxDllp, rxSkipNum, txTlp;
	UINT64 txTlpNum, txDllp, txSkipNum, xdWrBytes, xdRdBytes;
	UINT64 wrChanBytes[8] = {0};
	UINT64 rdChanBytes[8] = {0};
	BOOLEAN ret;

	DTRACE("ENTER\n");

	/* Get conters' value of the monitor */
	ret = PerfGetData(pperfRegs, &time, &tick, &rxTlp, &rxTlpNum, &rxDllp,
	    &rxSkipNum, &txTlp, &txTlpNum, &txDllp, &txSkipNum, wrChanBytes,
	    rdChanBytes, &xdWrBytes, &xdRdBytes);

	/* See if the request is successful. */
	if (ret) {
		DPRINTF(pcmd->ResMsg, MAX_MSG_SIZE,
			"T: %llu %u\n"
			"RX: %llu %llu %llu %llu\n"
			"TX: %llu %llu %llu %llu\n"
			"W03: %llu %llu %llu %llu\n"
			"W47: %llu %llu %llu %llu\n"
			"R03: %llu %llu %llu %llu\n"
			"R47: %llu %llu %llu %llu\n"
			"XD: %llu %llu\n",
			time, tick,
			rxTlp, rxTlpNum, rxDllp, rxSkipNum,
			txTlp, txTlpNum, txDllp, txSkipNum,
			wrChanBytes[0], wrChanBytes[1], wrChanBytes[2], wrChanBytes[3],
			wrChanBytes[4], wrChanBytes[5], wrChanBytes[6], wrChanBytes[7],
			rdChanBytes[0], rdChanBytes[1], rdChanBytes[2], rdChanBytes[3],
			rdChanBytes[4], rdChanBytes[5], rdChanBytes[6], rdChanBytes[7],
			xdWrBytes, xdRdBytes);

	} else {
		DEPRINTF(pcmd->ResMsg, MAX_MSG_SIZE,
			 "Failed to read performance registers.\n");
		status = STATUS_UNSUCCESSFUL;
	}

	return status;
}

/**
 * Handler of DMA_CMD_XPERF IOCTL command.
 *
 * This function calculates performance.
 *
 * @param[in] pDevExt   The device extension object.
 * @param[in,out] pcmd  The command object containing all arguments of requested
 *                      command. Returning messages are also stored in this
 *                      object.
 * @return NTSTATUS value to indicate success or failure.
 */
static UINT32 IoctlXPerf(PDEVICE_EXTENSION pDevExt, PCMD_INFO pcmd)
{
	PPERF_REGS pperfRegs = (PPERF_REGS)pDevExt->DmaRsc.DmaRegs;
	UCHAR tick;
	UCHAR tick_2nd;
	UINT64 time, rxTlp, rxTlpNum, rxDllp, rxSkipNum, txTlp, txTlpNum;
	UINT64 txDllp, txSkipNum, xdWrBytes, xdRdBytes;
	UINT64 time_2nd, rxTlp_2nd, rxTlpNum_2nd, rxDllp_2nd, rxSkipNum_2nd;
	UINT64 txTlp_2nd, txTlpNum_2nd, txDllp_2nd, txSkipNum_2nd;
	UINT64 xdWrBytes_2nd, xdRdBytes_2nd;
	UINT64 wrChanBytes[8] = {0};
	UINT64 rdChanBytes[8] = {0};
	UINT64 wrChanBytes_2nd[8] = {0};
	UINT64 rdChanBytes_2nd[8] = {0};
	UINT64 time_delta;
	UINT64 xdWrRate, xdRdRate;
	BOOLEAN ret;

	/* Get first sample of the performance registers */
	ret = PerfGetData(pperfRegs, &time, &tick, &rxTlp, &rxTlpNum, &rxDllp,
			  &rxSkipNum, &txTlp, &txTlpNum, &txDllp, &txSkipNum,
			  wrChanBytes, rdChanBytes, &xdWrBytes, &xdRdBytes);

	mdelay(100);

	/* See if the request is successful. */
	if (ret) {
		/* get 2nd sample */
		ret = PerfGetData(pperfRegs, &time_2nd, &tick_2nd, &rxTlp_2nd,
		&rxTlpNum_2nd, &rxDllp_2nd, &rxSkipNum_2nd, &txTlp_2nd,
		&txTlpNum_2nd, &txDllp_2nd, &txSkipNum_2nd, wrChanBytes_2nd,
		rdChanBytes_2nd, &xdWrBytes_2nd, &xdRdBytes_2nd);

		/* See if the request is successful. */
		if (ret) {
			time_delta = jiffies_to_nsecs(time_2nd-time);

			xdWrRate = (xdWrBytes_2nd - xdWrBytes);
			xdWrRate *= (1000 * 1000 * 1000);
			xdWrRate /= (1024 * 1024 * time_delta);

			xdRdRate = (xdRdBytes_2nd - xdRdBytes);
			xdRdRate *= (1000 * 1000 * 1000);
			xdRdRate /= (1024 * 1024 * time_delta);

			DPRINTF(pcmd->ResMsg, MAX_MSG_SIZE,
				"T: %llu \n"
				"XD: Read Rate=%llu MB/S Write Rate=%llu MB/S\n",
				time, xdRdRate, xdWrRate);

			return STATUS_SUCCESS;
		}
	}

	DEPRINTF(pcmd->ResMsg, MAX_MSG_SIZE,
		 "Failed to read performance registers.\n");

	return STATUS_UNSUCCESSFUL;
}

/**
 * Handler of DMA_CMD_MDUMP IOCTL command.
 *
 * This function doesn't call any HAL API. It dumps the data memory used by DMA
 * transfer. It can only dump 16 DWORD each time due to the message buffer
 * limit.
 *
 * @param[in] pDevExt   The device extension object.
 * @param[in,out] pcmd  The command object containing all arguments of requested
 *                      command. Returning messages are also stored in this
 *                      object.
 * @return NTSTATUS value to indicate success or failure.
 */
static UINT32 IoctlMemDump(PDEVICE_EXTENSION pDevExt, PCMD_INFO pcmd)
{

	UINT32 status = STATUS_SUCCESS;
	UINT8 chan = pcmd->Chan;
	UINT32 offset = pcmd->Offset;
	DMA_CHAN_END end = pcmd->End;
	DMA_DIRC dirc = pcmd->Dirc;
	UINT32 dout[16];
	PUINT32 addr;

	DTRACE("ENTER\n");

	DTRACE("Mem dump. channel=%d offset=%d len=%d\n", chan, offset, 16);

	/* Check if the offset is reasonable */
	offset &= 0xfffffffc;
	offset = (offset >= (pDevExt->ChunkLen - 16) * 4)
	    ? (pDevExt->ChunkLen - 16) * 4 : offset;

	/* Check if DDR mode is supported by the endpoint. */
	if (end == DMA_CHAN_DDR &&
	    (pDevExt->ChunkLen != pDevExt->RamSize / 64 / 4)) {
		DEPRINTF(pcmd->ResMsg, MAX_MSG_SIZE,
			"DDR memory is not available on EP.\n");
		status = STATUS_UNSUCCESSFUL;
		goto CLEANUP;
	}

	/* Get the data address of given channel and end */
	addr = GetBaseDataVAddr(pDevExt, dirc, chan, end);
	if (addr == NULL) {
		DEPRINTF(pcmd->ResMsg, MAX_MSG_SIZE,
			"Failed to get memory address.\n");
		status = STATUS_UNSUCCESSFUL;
		goto CLEANUP;
	}
	addr += offset / 4;

	/* Two different methods are used to access buffer data according to the
	 * location of the memory.
	 */
	if (end == DMA_CHAN_HST) {
		DumpSysMem(addr, 16, dout);
	} else {
		DumpLocRam(addr, 16, dout);
	}

	DPRINTF(pcmd->ResMsg, MAX_MSG_SIZE,
	    "[0x%06X]  %08x  %08x  %08x  %08x\n"
	    "[0x%06X]  %08x  %08x  %08x  %08x\n"
	    "[0x%06X]  %08x  %08x  %08x  %08x\n"
	    "[0x%06X]  %08x  %08x  %08x  %08x\n",
	    (offset + 0), dout[0], dout[1], dout[2], dout[3],
	    (offset + 16), dout[4], dout[5], dout[6], dout[7],
	    (offset + 32), dout[8], dout[9], dout[10], dout[11],
	    (offset + 48), dout[12], dout[13], dout[14], dout[15]);
CLEANUP:

	return status;
}



/**
 * Handler for DMA_CMD_STOP IOCTL command.
 *
 * This function calls HAL stop API to stop a transferring DMA channel. If the
 * channel is not running, this function takes no effect.
 *
 * @param[in] pDevExt   The device extension object.
 * @param[in,out] pcmd  The command object containing all arguments of requested
 *                      command. Returning messages are also stored in this
 *                      object.
 * @return NTSTATUS value to indicate success or failure.
 * @see HalDmaStop
 * @ingroup hal_caller
 */
static UINT32 IoctlDmaStop(PDEVICE_EXTENSION pDevExt, PCMD_INFO pcmd)
{
	UINT32 status = STATUS_SUCCESS;
	HAL_DMA_RET ret;

	DTRACE("ENTER\n");

	/* Call HAL stop API to stop the undergoing transfer on given channel. */
	ret = HalDmaStop(&pDevExt->DmaRsc, pcmd->Dirc, pcmd->Chan);

	/* See if the request is successful. */
	if (ret == DMA_SUCCESS) {
		DPRINTF(pcmd->ResMsg, MAX_MSG_SIZE,
			"OK! DMA %s on channel %d is stopped.\n",
			pcmd->Dirc == DMA_WRITE ? "write" : "read", pcmd->Chan);
	} else {
		DEPRINTF(pcmd->ResMsg, MAX_MSG_SIZE,
			 "Failed to stop the transfer on given channel:\n%s\n",
			 HalDmaGetErrInfo(ret));
			 status = STATUS_UNSUCCESSFUL;
	}

	return status;
}


/**
 * Handler for DMA_CMD_RESUME IOCTL command.
 *
 * This function calls HAL resume API to resume a stopped DMA channel. If the
 * channel is running, this function takes no effect.
 *
 * @param[in] pDevExt   The device extension object.
 * @param[in,out] pcmd  The command object containing all arguments of requested
 *                      command. Returning messages are also stored in this
 *                      object.
 * @return NTSTATUS value to indicate success or failure.
 * @see HalDmaResume
 * @ingroup hal_caller
 */
static UINT32 IoctlDmaResume(PDEVICE_EXTENSION pDevExt, PCMD_INFO pcmd)
{
	UINT32 status = STATUS_SUCCESS;
	HAL_DMA_RET ret;

	DTRACE("ENTER\n");

	/* Call HAL resume API to stop the undergoing transfer on given channel. */
	ret = HalDmaResume(&pDevExt->DmaRsc, pcmd->Dirc, pcmd->Chan);

	/* See if the request is successful. */
	if (ret == DMA_SUCCESS) {
		DPRINTF(pcmd->ResMsg, MAX_MSG_SIZE,
		"OK! DMA %s on channel %d is resumed.\n",
		pcmd->Dirc == DMA_WRITE ? "write" : "read", pcmd->Chan);
	} else {
		DEPRINTF(pcmd->ResMsg, MAX_MSG_SIZE,
		"Failed to resume the transfer on given channel:\n%s\n",
		HalDmaGetErrInfo(ret));
		status = STATUS_UNSUCCESSFUL;
	}

	return status;
}


/**
 * Handler for DMA_CMD_ABORT IOCTL command.
 *
 * This function calls HAL abort API to stop a transferring DMA channel. If the
 * channel is not running, this function takes no effect.
 *
 * @param[in] pDevExt   The device extension object.
 * @param[in,out] pcmd  The command object containing all arguments of requested
 *                      command. Returning messages are also stored in this
 *                      object.
 * @return NTSTATUS value to indicate success or failure.
 * @see HalDmaAbort
 * @ingroup hal_caller
 */
static UINT32 IoctlDmaAbort(PDEVICE_EXTENSION pDevExt, PCMD_INFO pcmd)
{
	UINT32 status = STATUS_SUCCESS;
	HAL_DMA_RET ret;

	DTRACE("ENTER\n");

	/* Call HAL abort API to abort the undergoing transfer on given channel. */
	ret = HalDmaAbort(&pDevExt->DmaRsc, pcmd->Dirc, pcmd->Chan);

	/* See if the request is successful. */
	if (ret == DMA_SUCCESS) {
		DPRINTF(pcmd->ResMsg, MAX_MSG_SIZE,
			"OK! DMA %s on channel %d is aborted.\n",
			pcmd->Dirc == DMA_WRITE ? "write" : "read", pcmd->Chan);
	} else {
		DEPRINTF(pcmd->ResMsg, MAX_MSG_SIZE,
			 "Failed to abort the transfer on given channel:\n%s\n",
			 HalDmaGetErrInfo(ret));
			 status = STATUS_UNSUCCESSFUL;
	}

	return status;
}


/**
 * Handler for DMA_CMD_DW and DMA_CMD_DR IOCTL commands.
 *
 * This function firstly checks the current state of given DMA channel to make
 * sure it's not running, then generates data pattern in source buffer, finally
 * calls corresponding HAL R/W API to start the transfer. Note that there are
 * four DMA R/W APIs and you should call the correct version based on the
 * requests.
 *
 * @param[in] pDevExt   The device extension object.
 * @param[in,out] pcmd  The command object containing all arguments of requested
 *                      command. Returning messages are also stored in this
 *                      object.
 * @return NTSTATUS value to indicate success or failure.
 * @see HalDmaState
 * @see HalDmaSingleBlockWrite
 * @see HalDmaSingleBlockRead
 * @see HalDmaMultiBlockWrite
 * @see HalDmaMultiBlockRead
 * @ingroup hal_caller
 */
static UINT32 IoctlDmaRW(PDEVICE_EXTENSION pDevExt, PCMD_INFO pcmd)
{
	UINT32 status = STATUS_SUCCESS;
	HAL_DMA_RET ret;
	PDMA_RSC pdrsc = &pDevExt->DmaRsc;
	DMA_DIRC dirc = pcmd->Cmd == DMA_CMD_DW ? DMA_WRITE : DMA_READ;
	UINT8 lle = pcmd->LLElements;
	UINT8 chan = pcmd->Chan;
	UINT32 blkSize = pcmd->Len * 4;
	BOOLEAN queue = pcmd->Queue;
	BOOLEAN isDdr = pcmd->Ddr;
	UINT8 wm = pcmd->WaterMark;
	UINT8 wei = pcmd->Weight - 1;   // actual weights are 0 to 31
	UINT32 recyc = pcmd->Recyc;
	BOOLEAN chk = pcmd->Chk;
	DMA_CHAN_STATUS stat;
	PCHAN_INFO_EX pChInfEx;
	static BOOLEAN flip = TRUE;

	DTRACE("ENTER\n");

	/* Check transfer size. More strict than HAL check. */
	if (blkSize > 4 * pDevExt->ChunkLen) {
		DEPRINTF(pcmd->ResMsg, MAX_MSG_SIZE,
			 "Transfer block length is invalid: %d > %d.\n",
			 blkSize / 4, pDevExt->ChunkLen);
		status = STATUS_UNSUCCESSFUL;
		goto CLEANUP;
	}

	pChInfEx = (dirc == DMA_WRITE) ? &pDevExt->WrChInfEx[chan]
				       : &pDevExt->RdChInfEx[chan];

	/* Check channel state to make sure it's not running, otherwise we can't
	* start a new transfer.
	*/
	ret = HalDmaState(pdrsc, dirc, chan, &stat, NULL, NULL, NULL, NULL);
	if (ret != DMA_SUCCESS) {
		DEPRINTF(pcmd->ResMsg, MAX_MSG_SIZE,
			 "Failed to get DMA channel state:\n%s\n",
			 HalDmaGetErrInfo(ret));
			 status = STATUS_UNSUCCESSFUL;
		goto CLEANUP;
	}
	if (stat == CHAN_RUNNING) {
		DEPRINTF(pcmd->ResMsg, MAX_MSG_SIZE,
			 "DMA %s channel %d is still running.\n",
			 dirc == DMA_WRITE ? "write" : "read", chan);
		status = STATUS_UNSUCCESSFUL;
		goto CLEANUP;
	}

	/* Check if DDR mode is supported by the endpoint. */
	if (isDdr && (pDevExt->ChunkLen != pDevExt->RamSize / 64 / 4)) {
		DEPRINTF(pcmd->ResMsg, MAX_MSG_SIZE,
			 "DDR memory is not available on EP.\n");
		status = STATUS_UNSUCCESSFUL;
		goto CLEANUP;
	}

	/* Prepare data pattern in source buffer. After DMA transfer get finished,
	* we can use this pattern to check if the data is correctly copied to
	* desitnation.
	*/
	flip = !flip;
	pChInfEx->BlockSize = blkSize;
	pChInfEx->Pattern = DEFAULT_PATTERN + (chan << 28);
	pChInfEx->PatternIncre = isDdr ? (flip ? 3 : 4) : (flip ? 1 : 2);
	pChInfEx->DataCheck = chk ? DATA_CHECK_UNKNOWN : DATA_CHECK_NONE;
	pChInfEx->TotalSize = lle == 0 ? blkSize : (UINT64) blkSize * (lle - 1)
	    * recyc;
	GenSrcData(pDevExt, dirc, chan, pChInfEx->Pattern,
	    pChInfEx->PatternIncre, isDdr);

	pChInfEx->tick_count_start = jiffies;

	if (lle == 0) {
		/* Single block transfer */
		PHYADDR src = GetBaseDataPAddr(pDevExt, dirc, chan,
		GET_CHAN_SRC(dirc, isDdr));
		PHYADDR dst = GetBaseDataPAddr(pDevExt, dirc, chan,
		GET_CHAN_DST(dirc, isDdr));
		if (dirc == DMA_WRITE)
			ret = HalDmaSingleBlockWrite(pdrsc, chan, src, dst, blkSize, wei, queue);
		else
			ret = HalDmaSingleBlockRead(pdrsc, chan, src, dst, blkSize, wei, queue);
	} else {
		/* Multi block transfer */
		wm = wm == 0 ? (lle / 2 - 1) : wm;
		if (dirc == DMA_WRITE)
			ret = HalDmaMultiBlockWrite(pdrsc, chan, lle, wm, recyc, wei, queue,
						    PeddProducerFunc, pDevExt);
		else
			ret = HalDmaMultiBlockRead(pdrsc, chan, lle, wm, recyc, wei, queue,
						   PeddProducerFunc, pDevExt);
	}

	/* Transfer request is submitted, and we need to check the return value to
	* see if everything goes fine.
	*/
	if (ret == DMA_SUCCESS) {
		if (queue)
		    DPRINTF(pcmd->ResMsg, MAX_MSG_SIZE,
			    "OK! DMA %s on channel %d is queued up.\n",
			    dirc == DMA_WRITE ? "write" : "read", chan);
		else
		    DPRINTF(pcmd->ResMsg, MAX_MSG_SIZE,
			    "OK! DMA %s on channel %d is started.\n",
			    dirc == DMA_WRITE ? "write" : "read", chan);
	} else {
		DEPRINTF(pcmd->ResMsg, MAX_MSG_SIZE,
			"Failed to start DMA %s:\n%s\n",
			dirc == DMA_WRITE ? "write" : "read",
			HalDmaGetErrInfo(ret));
			status = STATUS_UNSUCCESSFUL;
	}

CLEANUP:

	return status;
}

/**
 * Handler for DMA_CMD_STATE IOCTL command.
 *
 * This function calls HAL state API to get current state of given channel.
 * It also adds additional information like the transferred size and data check
 * result.
 *
 * @param[in] pDevExt   The device extension object.
 * @param[in,out] pcmd  The command object containing all arguments of requested
 *                      command. Returning messages are also stored in this
 *                      object.
 * @return NTSTATUS value to indicate success or failure.
 * @see HalDmaState
 * @ingroup hal_caller
 */
static UINT32 IoctlDmaState(PDEVICE_EXTENSION pDevExt, PCMD_INFO pcmd)
{
	UINT32 status = STATUS_SUCCESS;
	HAL_DMA_RET ret;
	PDMA_RSC pdrsc = &pDevExt->DmaRsc;
	DMA_DIRC dirc = pcmd->Dirc;
	UINT8 chan = pcmd->Chan;
	DMA_CHAN_STATUS stat;
	DMA_ERR err;
	UINT64 xferred = 0;
	UINT32 intCntDone = 0;
	UINT32 intCntAbort = 0;
	PCHAN_INFO_EX pChInfEx;
	UINT64 ticks;
	UINT64 msecs;
	UINT64 bpermsec;
	PCHAR statusStr;
	PCHAR errStr;
	PCHAR dcStr;

	DTRACE("ENTER\n");

	/* Call HAL API to get current state of given channel. */
	ret = HalDmaState(pdrsc, dirc, chan, &stat, &err, &xferred, &intCntDone,
	    &intCntAbort);

	if (ret != DMA_SUCCESS) {
		DEPRINTF(pcmd->ResMsg, MAX_MSG_SIZE,
			"Failed to get DMA channel state:\n%s\n",
			HalDmaGetErrInfo(ret));
		status = STATUS_FAIL;
	} else {
		pChInfEx = (dirc == DMA_WRITE) ? &pDevExt->WrChInfEx[chan]
					       : &pDevExt->RdChInfEx[chan];

		if (stat == CHAN_RUNNING)
			pChInfEx->tick_count_stop = jiffies;

		ticks = pChInfEx->tick_count_stop - pChInfEx->tick_count_start;
		msecs = jiffies_to_msecs(ticks);

		msecs = msecs ? msecs : 1;
		bpermsec = PEDD_DIV64(PEDD_DIV64(xferred * 1000, 1024), msecs);

		statusStr = stat == CHAN_RUNNING ? "Running"
			: stat == CHAN_STOPPED ? "Stopped"
			: stat == CHAN_HALTED ? "Halted"
			: stat == CHAN_QUEUING ? "Queuing"
			: "Unknown";

		errStr = err == DMA_ERR_WR ? "Application Write Error"
			: err == DMA_ERR_RD ? "Application Read Error"
			: err == DMA_ERR_FETCH_LL ? "Linked List Element Fetch Error"
			: err == DMA_ERR_UNSUPPORTED_RQST ? "Unsupported Request"
			: err == DMA_ERR_COMPLETER_ABORT ? "Completer Abort"
			: err == DMA_ERR_CPL_TIME_OUT ? "CPL Time Out"
			: err == DMA_ERR_DATA_POISONING ? "Data Poisoning"
			: "None";

		dcStr = pChInfEx->DataCheck == DATA_CHECK_NONE ? "N/A"
			: pChInfEx->DataCheck == DATA_CHECK_FAILURE ? "Failure"
			: pChInfEx->DataCheck == DATA_CHECK_SUCCESS ? "Success"
			: "Unknown";

		DPRINTF(pcmd->ResMsg, MAX_MSG_SIZE,
			"Channel:      %s %d\n"
			"State:        %s\n"
			"HW Error:     %s\n"
			"Transferred:  %llu (%llu bytes in total)\n"
			"Time Elapsed: %llu msec\n"
			"Rate:         %llu KB/s\n"
			"Interrupts:   %d(Done)  %d(Abort)\n"
			"Data check:   %s\n",
			dirc == DMA_WRITE ? "Write" : "Read", chan,
			statusStr,
			errStr,
			xferred, pChInfEx->TotalSize,
			msecs,
			bpermsec,
			intCntDone, intCntAbort, dcStr);
	}

	return status;
}

/**
 * Handler for DMA_CMD_INIT IOCTL command.
 *
 * This function allocates the system buffers used by DMA transfer as source
 * and destination, and calls HAL initialization API to set MSI address and
 * data. If the DMA core works in loop back mode, an extra loop back system
 * buffers is allocated too.
 *
 * @param[in,out] pDevExt    The device extension object.
 * @param[in,out] pcmd       The command object containing all arguments of
 *                          requested command. Returning messages are also
 *                          stored in this object.
 * @return NTSTATUS value to indicate success or failure.
 * @see HalDmaInit
 * @ingroup hal_caller
 */
static UINT32 IoctlDmaInit(struct pedd_dev *priv, PDEVICE_EXTENSION pDevExt, PCMD_INFO pcmd)
{
	UINT32 status = STATUS_SUCCESS;
	UINT32 i;
	HAL_DMA_RET ret;
	UINT32 ddrStatus;
	UINT32 llSlots[4] = {CHUNK_ID_LLA, CHUNK_ID_LLB, CHUNK_ID_LLC, CHUNK_ID_LLD};
	XDATA_RET xret;

	DTRACE("ENTER\n");

	if (eDMA) {
		struct msi_msg msg;

		if (pDevExt->ChunkLen) {
			status = STATUS_FAIL;
			DEPRINTF(pcmd->ResMsg, MAX_MSG_SIZE,
				 "Device was already initialized!\n");
			goto CLEANUP;
		}
	
		if (priv->pdev->msi_enabled)
			INFO_PRINT("IRQ MSI selected\n");
		else if (priv->pdev->msix_enabled)
			INFO_PRINT("IRQ MSI-X selected\n");

		get_cached_msi_msg(pci_irq_vector(priv->pdev, 0), &msg);

		pDevExt->DmaRsc.MsiAddr.HighPart = msg.address_hi;
		pDevExt->DmaRsc.MsiAddr.LowPart = msg.address_lo;
		pDevExt->DmaRsc.MsiData = msg.data;
		INFO_PRINT("DmaRsc.MsiAddr_h=0x%.8x DmaRsc.MsiAddr_l=0x%.8x DmaRsc.MsiData=0x%.8x\n",
			   pDevExt->DmaRsc.MsiAddr.HighPart, pDevExt->DmaRsc.MsiAddr.LowPart,
			   pDevExt->DmaRsc.MsiData);

		/* Call HAL init function to reset DMA logic and renew MSI settings.
		 * We can also get the number of enabled write/read channels, which is
		 * used to segment the memory.
		 */
		ret = HalDmaInit(&pDevExt->DmaRsc);
		if (ret != DMA_SUCCESS)	{
			DEPRINTF(pcmd->ResMsg, MAX_MSG_SIZE,
				 "Failed to initial eDMA device:\n%s\n",
				 HalDmaGetErrInfo(ret));
			status = STATUS_FAIL;
			goto CLEANUP;
		}
	}

	/* Calculate IO memory size.
	 * 1) When DDR mode is enabled
	 * The highest 6 bits of available RamSize are used as chunk id, so the
	 * size of each chunk is RamSize/(2^6). That's to say there are 64 chunks.
	 * 2) When DDR mode is NOT enabled
	 * The highest 5 bits of available RamSize are used as chunk id, so the
	 * size of each chunk is RamSize/(2^5). That's to say there are 32 chunks.
	 *
	 * In both cases, we at least use 21 chunks: 4 chunks for linked list space
	 * (local RAM only), 16 chunks for DMA channels, and 1 chunk for xData.
	 *
	 * But for 1), we also use 2 additional chunks started from the second half
	 * of the space: 1 chunk for DDR W0 and 1 chunk for DDR R0.
	 * See CHUNK_ID enum.
	 */
	ddrStatus = PEDD_READ_UINT32((PUINT8)priv->bar[xData_bar] + 0x24);
	if ((ddrStatus >> 31))
		pDevExt->ChunkLen = pDevExt->RamSize / 64 / 4;
	else
		pDevExt->ChunkLen = pDevExt->RamSize / 32 / 4;

	INFO_PRINT("ChunkLen=%d RamSize=%d\n", pDevExt->ChunkLen, pDevExt->RamSize);

	if (eDMA) {
		INFO_PRINT("DmaRsc.WrChans=%d DmaRsc.RdChans=%d\n", pDevExt->DmaRsc.WrChans, pDevExt->DmaRsc.RdChans);

		status = BatchAllocDmaMem(priv, pDevExt->DmaRsc.WrChans, pDevExt->DmaRsc.RdChans);

		if (status != STATUS_SUCCESS) {
			DEPRINTF(pcmd->ResMsg, MAX_MSG_SIZE,
				"Failed to allocate continuous system memory for DMA!\n");
			goto CLEANUP;
		}
		INFO_PRINT("Allocated memory for 2rd, 2wr channels, now build link list memory \n");
		/* Set linked list addresses */
		for (i = 0; i < 16; i++) {
			if (i % 4 == 0) {
				pDevExt->DmaRsc.LLPhyAddr[i].QuadPart =
					pDevExt->RamPa.QuadPart +
					llSlots[i/4] * pDevExt->ChunkLen * 4;
				pDevExt->DmaRsc.LLVirAddr[i] =
					pDevExt->RamVa + llSlots[i/4] * pDevExt->ChunkLen;
			} else {
				pDevExt->DmaRsc.LLPhyAddr[i].QuadPart =
					pDevExt->DmaRsc.LLPhyAddr[i-1].QuadPart + MAX_LL_LEN * 4;
				pDevExt->DmaRsc.LLVirAddr[i] =
					pDevExt->DmaRsc.LLVirAddr[i-1] + MAX_LL_LEN;
			}
		}
	}
	INFO_PRINT("Setup Callback function to PeddXferDoneCallback\n");

	/* Set xfer done callback */
	pDevExt->DmaRsc.XferDoneFunc = PeddXferDoneCallback;
	pDevExt->DmaRsc.XferDoneContext = pDevExt;

	/* xData init */
	if (xData) {
		pDevExt->xDataRsc.xDataLocMemSize = pDevExt->ChunkLen * 4;
		pDevExt->xDataRsc.xDataLocMemPa.QuadPart = pDevExt->RamPa.QuadPart +
			CHUNK_ID_XDATA * pDevExt->ChunkLen * 4;
		pDevExt->xDataRsc.xDataLocMemVa = pDevExt->RamVa + CHUNK_ID_XDATA
			* pDevExt->ChunkLen;
		pDevExt->xDataRsc.xDataMemSize = XDATA_BUF_SIZE;

		status = AllocProtocolMem(priv, XDATA_BUF_SIZE, &pDevExt->xDataRsc.xDataMemVa, &pDevExt->xDataRsc.xDataBus, &pDevExt->xDataRsc.xDataMemPa);

		if (status != STATUS_SUCCESS) {
			DEPRINTF(pcmd->ResMsg, MAX_MSG_SIZE,
				"Failed to allocate contiguous system memory for XDATA!\n");
			goto CLEANUP;
		}
		xret = xDataInit(&pDevExt->xDataRsc, pDevExt->ReadMTU, pDevExt->WriteMTU);

		if (xret != XDATA_SUCCESS) {
			DEPRINTF(pcmd->ResMsg, MAX_MSG_SIZE,
				 "Failed to initialize xData engine\n");
			status = STATUS_FAIL;
			goto CLEANUP;
		}
	}
	else
		INFO_PRINT("No xData. Done with IoctlDmaInit\n");

	DPRINTF(pcmd->ResMsg, MAX_MSG_SIZE, "OK! PEDD is initialized.\n");

CLEANUP:
	return status;
}


/**
 * IOCTL commands dispatcher.
 *
 * This function calls appropriate sub routines to server given IOCTL command.
 *
 * @param[in] pDevExt   The device extension object.
 * @param[in] pIrp      The IRP object.
 * @param[in,out] pcmd  The command object containing all arguments of requested
 *                      command. Returning messages are also stored in this
 *                      object.
 * @return NTSTATUS value to indicate success or failure.
 */
static UINT32 IoctlCommand(struct pedd_dev *priv, PDEVICE_EXTENSION pDevExt, PCMD_INFO pcmd)
{
	UINT32 status = STATUS_FAIL;
	DMA_CMD cmd = pcmd->Cmd;

	pcmd->ResMsg[0] = '\0';

	INFO_PRINT("+++++++++++++++++  PEDD_IOCTL_APP_CMD - [%d] +++++++++++++++++", cmd);

	/* All commands except "init" itself can't be processed if DMA is not
	 * yet initialized.
	 */
	if ((eDMA || xData) && cmd != DMA_CMD_INIT && pDevExt->ChunkLen == 0) {
		DEPRINTF(pcmd->ResMsg, MAX_MSG_SIZE, "DMA not initialized!\n");
		return STATUS_FAIL;
	}
	switch (cmd) {
	case DMA_CMD_INIT:
		if (eDMA || xData)
			status = IoctlDmaInit(priv, pDevExt, pcmd);
		break;
	case DMA_CMD_STATE:
		if (eDMA)
			status = IoctlDmaState(pDevExt, pcmd);
		break;
	case DMA_CMD_DR:
	case DMA_CMD_DW:
		if (eDMA)
			status = IoctlDmaRW(pDevExt, pcmd);
		break;
	case DMA_CMD_ABORT:
		if (eDMA)
			status = IoctlDmaAbort(pDevExt, pcmd);
		break;
	case DMA_CMD_STOP:
		if (eDMA)
			status = IoctlDmaStop(pDevExt, pcmd);
		break;
	case DMA_CMD_RESUME:
		if (eDMA)
			status = IoctlDmaResume(pDevExt, pcmd);
		break;
	case DMA_CMD_DQF:
		if (eDMA)
			status = IoctlDmaFlushQueue(pDevExt, pcmd);
		break;
	case DMA_CMD_MDUMP:
		if (eDMA)
			status = IoctlMemDump(pDevExt, pcmd);
		break;
	case DMA_CMD_MADDR:
		if (eDMA)
			status = IoctlBufAddr(pDevExt, pcmd);
		break;
	case DMA_CMD_RDUMP:
		if (eDMA)
			status = IoctlDmaRegsDump(pDevExt, pcmd);
		break;
	case DMA_CMD_LLDUMP:
		if (eDMA)
			status = IoctlDmaLLDump(pDevExt, pcmd);
		break;
	case DMA_CMD_PERF:
		if (eDMA)
			status = IoctlPerf(pDevExt, pcmd);
		break;
	case DMA_CMD_XPERF:
		if (xData)
			status = IoctlXPerf(pDevExt, pcmd);
		break;
	case DMA_CMD_XSTART:
		if (xData)
			status = IoctlxDataStartXfer(pDevExt, pcmd);
		break;
	case DMA_CMD_XSTOP:
		if (xData)
			status = IoctlxDataStopXfer(pDevExt, pcmd);
		break;
	case DMA_CMD_XSTATUS:
		if (xData)
			status = IoctlxDataStatus(pDevExt, pcmd);
		break;
	default:
		DEPRINTF(pcmd->ResMsg, MAX_MSG_SIZE, "Unknown command");
		break;
	}

	return status;
}

/**
 * Parse command string passed from user space.
 *
 * @param[in] msg       Command message.
 * @param[out] cmdinfo  Command info object holding various command arguments.
 * @return True if the command message is parsed successfully; False otherwise.
 */
static BOOLEAN ParseCommand(CHAR *msg, PCMD_INFO cmdinfo)
{

	ULONG args[PEDD_CB_NUM_ARGS];
	CHAR command[64];

	/* Call pedd command parser */
	if (pedd_parse_cmd(msg, command, args)) {
		// Parse Error
		DEPRINTF(cmdinfo->ResMsg, MAX_MSG_SIZE, "Failed to parse the command.\n");
		return FALSE;
	}

	if (!strcmp(command, "init"))
		cmdinfo->Cmd = DMA_CMD_INIT;
	else if (!strcmp(command, "state"))
		cmdinfo->Cmd = DMA_CMD_STATE;
	else if (!strcmp(command, "dw"))
		cmdinfo->Cmd = DMA_CMD_DW;
	else if (!strcmp(command, "dr"))
		cmdinfo->Cmd = DMA_CMD_DR;
	else if (!strcmp(command, "dqf"))
		cmdinfo->Cmd = DMA_CMD_DQF;
	else if (!strcmp(command, "abort"))
		cmdinfo->Cmd = DMA_CMD_ABORT;
	else if (!strcmp(command, "stop"))
		cmdinfo->Cmd = DMA_CMD_STOP;
	else if (!strcmp(command, "resume"))
		cmdinfo->Cmd = DMA_CMD_RESUME;
	else if (!strcmp(command, "mdump"))
		cmdinfo->Cmd = DMA_CMD_MDUMP;
	else if (!strcmp(command, "maddr"))
		cmdinfo->Cmd = DMA_CMD_MADDR;
	else if (!strcmp(command, "rdump"))
		cmdinfo->Cmd = DMA_CMD_RDUMP;
	else if (!strcmp(command, "lldump"))
		cmdinfo->Cmd = DMA_CMD_LLDUMP;
	else if (!strcmp(command, "dperf"))
		cmdinfo->Cmd = DMA_CMD_PERF;
	else if (!strcmp(command, "xperf"))
		cmdinfo->Cmd = DMA_CMD_XPERF;
	else if (!strcmp(command, "xstart"))
		cmdinfo->Cmd = DMA_CMD_XSTART;
	else if (!strcmp(command, "xstop"))
		cmdinfo->Cmd = DMA_CMD_XSTOP;
	else if (!strcmp(command, "xstatus"))
		cmdinfo->Cmd = DMA_CMD_XSTATUS;
	else {
		DEPRINTF(cmdinfo->ResMsg, MAX_MSG_SIZE, "Unknown command: %s\n", command);
		return FALSE;
	}

	cmdinfo->Chan = (UINT8)args[PEDD_CB_CHAN_IDX];
	cmdinfo->Dirc = (DMA_DIRC)args[PEDD_CB_DIRC_IDX];
	cmdinfo->End = (DMA_CHAN_END)args[PEDD_CB_END_IDX];
	cmdinfo->LLElements = (UINT8)args[PEDD_CB_LLE_IDX];
	cmdinfo->WaterMark = (UINT8)args[PEDD_CB_WM_IDX];
	cmdinfo->Weight = (UINT8)args[PEDD_CB_WEI_IDX];
	cmdinfo->Len = args[PEDD_CB_LEN_IDX];
	cmdinfo->Recyc = args[PEDD_CB_RECYC_IDX];
	cmdinfo->Offset = args[PEDD_CB_OFFSET_IDX];
	cmdinfo->Chk = (BOOLEAN)args[PEDD_CB_CHECK_IDX];
	cmdinfo->Queue = (BOOLEAN)args[PEDD_CB_QUEUE_IDX];
	cmdinfo->Ddr = (BOOLEAN)args[PEDD_CB_DDR_IDX];
	cmdinfo->Upstream = (BOOLEAN)args[PEDD_CB_UP_IDX];

	return TRUE;
}

void pedd_Read_Header(struct pci_dev *pdev, PCI_COMMON_CONFIG *data)
{
	pci_read_config_word(pdev, PCI_VENDOR_ID, &data->VendorID);
	pci_read_config_word(pdev, PCI_DEVICE_ID, &data->DeviceID);
	pci_read_config_word(pdev, PCI_COMMAND, &data->Command);
	pci_read_config_word(pdev, PCI_STATUS, &data->Status);
	pci_read_config_byte(pdev, PCI_REVISION_ID, &data->RevisionID);
	pci_read_config_byte(pdev, PCI_CLASS_PROG, &data->ProgIf);
	pci_read_config_byte(pdev, 0xa, &data->SubClass);
	pci_read_config_byte(pdev, 0xb, &data->BaseClass);
	pci_read_config_byte(pdev, PCI_CACHE_LINE_SIZE, &data->CacheLineSize);
	pci_read_config_byte(pdev, PCI_LATENCY_TIMER, &data->LatencyTimer);
	pci_read_config_byte(pdev, PCI_HEADER_TYPE, &data->HeaderType);
	pci_read_config_byte(pdev, PCI_BIST, &data->BIST);

	pci_read_config_dword(pdev, PCI_BASE_ADDRESS_0, (u32 *)&data->u.type0.BaseAddresses[0]);
	pci_read_config_dword(pdev, PCI_BASE_ADDRESS_1, (u32 *)&data->u.type0.BaseAddresses[1]);
	pci_read_config_dword(pdev, PCI_BASE_ADDRESS_2, (u32 *)&data->u.type0.BaseAddresses[2]);
	pci_read_config_dword(pdev, PCI_BASE_ADDRESS_3, (u32 *)&data->u.type0.BaseAddresses[3]);
	pci_read_config_dword(pdev, PCI_BASE_ADDRESS_4, (u32 *)&data->u.type0.BaseAddresses[4]);
	pci_read_config_dword(pdev, PCI_BASE_ADDRESS_5, (u32 *)&data->u.type0.BaseAddresses[5]);

	pci_read_config_word(pdev, PCI_SUBSYSTEM_VENDOR_ID, &data->u.type0.SubVendorID);
	pci_read_config_word(pdev, PCI_SUBSYSTEM_ID, &data->u.type0.SubSystemID);
	pci_read_config_dword(pdev, PCI_ROM_ADDRESS, (u32 *)&data->u.type0.ROMBaseAddress);
	pci_read_config_byte(pdev, PCI_CAPABILITY_LIST, &data->u.type0.CapabilitiesPtr);

	pci_read_config_byte(pdev, PCI_INTERRUPT_LINE, &data->u.type0.InterruptLine);
	pci_read_config_byte(pdev, PCI_INTERRUPT_PIN, &data->u.type0.InterruptPin);
	pci_read_config_byte(pdev, PCI_MIN_GNT, &data->u.type0.MinimumGrant);
	pci_read_config_byte(pdev, PCI_MAX_LAT, &data->u.type0.MaximumLatency);

}

void pedd_rw_pci_cfg(struct pci_dev *pdev, INT d, UINT32 length, UINT32 offset, UCHAR *val)
{
	UINT32 i;

	if (d == 0) {//0 for read, 1 for write
		switch (length)	{
		case 1:
			pci_read_config_byte(pdev, offset, val);
			break;
		case 2:
			pci_read_config_byte(pdev, offset, val);
			pci_read_config_byte(pdev, (offset+1), (val+1));
			break;
		case 4:
			pci_read_config_byte(pdev, offset, val);
			pci_read_config_byte(pdev, (offset+1), (val+1));
			pci_read_config_byte(pdev, (offset+2), (val+2));
			pci_read_config_byte(pdev, (offset+3), (val+3));
			break;
		default:
			for (i = 0; i < length; i++)
				pci_read_config_byte(pdev, (offset+i), (val+i));
		}
	} else {
		switch (length) {
		case 1:
			pci_write_config_byte(pdev, offset, *val);
			break;
		case 2:
			pci_write_config_byte(pdev, offset, *val);
			pci_write_config_byte(pdev, (offset+1), *(val+1));
			break;
		case 4:
			pci_write_config_byte(pdev, offset, *val);
			pci_write_config_byte(pdev, (offset+1), *(val+1));
			pci_write_config_byte(pdev, (offset+2), *(val+2));
			pci_write_config_byte(pdev, (offset+3), *(val+3));
			break;
		default:
			for (i = 0; i < length; i++)
				pci_write_config_byte(pdev, (offset+i), *(val+i));
		}
	}
}

void pedd_rw_pci_bar(INT d, UINT32 length, void *bar_addr, UINT32 offset, UCHAR *val)
{
	if (d == 0) { //0 for read, 1 for write
		switch (length) {
		case 1:
			*(PUINT8)val = PEDD_READ_UINT8((PCHAR)bar_addr + offset);
			break;
		case 2:
			*(PUINT16)val = PEDD_READ_UINT16((PCHAR)bar_addr + offset);
			break;
		case 4:
			*(PUINT32)val = PEDD_READ_UINT32((PCHAR)bar_addr + offset);
			break;
		default:
			ERROR_PRINT("wrong read length\n");
		}
	} else {
		switch (length) {
		case 1:
			PEDD_WRITE_UINT8((PCHAR)bar_addr + offset, *(PUINT8)val);
			break;
		case 2:
			PEDD_WRITE_UINT16((PCHAR)bar_addr + offset, *(PUINT16)val);
			break;
		case 4:
			PEDD_WRITE_UINT32((PCHAR)bar_addr + offset, *(PUINT32)val);
			INFO_PRINT("addr=0x%pP,val32=0x%x\n", (PCHAR)bar_addr + offset, *(PUINT32)val);
			break;
		default:
			ERROR_PRINT("wrong write length\n");
		}
	}
}

int pedd_rw_pci_hdr(struct pci_dev *pdev, INT rc, INT d, UINT32 length, UINT32 offset, UCHAR *val)
{
	struct pci_dev *rc_dev = pci_get_device(RC_SNPS_VENDOR_ID, RC_SNPS_DEVICE_ID, NULL);

	if (!rc) {
		pedd_rw_pci_cfg(pdev, d, length, offset, val);
		return STATUS_SUCCESS;
	}

	if (!rc_dev) {
		ERROR_PRINT("Can't find rc\n");
		return STATUS_UNSUCCESSFUL;
	}

	pedd_rw_pci_cfg(rc_dev, d, length, offset, val);
	return STATUS_SUCCESS;
}

static void pedd_unmap_bars(struct pedd_dev *priv, struct pci_dev *pdev)
{
	int i;

	for (i = 0; i < PEDD_BAR_NUM; i++) {
		/* is this BAR mapped? */
		if (priv->bar[i]) {
			/* unmap BAR */
			pci_iounmap(pdev, priv->bar[i]);
			priv->bar[i] = NULL;
		}
	}
}

static int pedd_map_bars(struct pedd_dev *priv, struct pci_dev *pdev)
{
	int rc = 0;
	int i;

	resource_size_t bar_start, bar_end, bar_length;

	/* iterate through all the BARs */
	for (i = 0; i < PEDD_BAR_NUM; i++) {
		bar_start = pci_resource_start(pdev, i);
		bar_end = pci_resource_end(pdev, i);

		priv->bar[i] = NULL;

		/* do not map BARs with address 0 */
		if (!bar_start || !bar_end) {
			rc = -1;
			continue;
		}
		//bar_length = bar_end - bar_start + 1;
		bar_length = pci_resource_len(pdev, i);

		/* map the device memory or IO region into kernel virtual address space */
		priv->bar[i] = pci_iomap(pdev, i, bar_length);
		priv->bar_len[i] = bar_length;
		priv->RegSpacePa[i].QuadPart = bar_start;
		priv->RegSpacePa[i].HighPart = pci_dma_h(bar_start);
		priv->RegSpacePa[i].LowPart = pci_dma_l(bar_start);
		if (!priv->bar[i]) {
			ERROR_PRINT("Could not map BAR #%d.\n", i);
			rc = -1;
			goto fail;
		}
		INFO_PRINT("BAR[%d] mapped at 0x%p with length %lu.\n", i, priv->bar[i], (unsigned long)bar_length);
	}

	/* succesfully mapped all required BAR regions */
	rc = 0;
	goto success;
fail:
	/* unmap any BARs that we did map */
	pedd_unmap_bars(priv, pdev);
success:
	return rc;
}
#if 0


static int transfer_build(struct xdma_transfer *transfer, u64 ep_addr,
		int dir_to_dev, int non_incr_addr, int force_new_desc,
		int userspace)
{
	int i = 0;
	int j = 0;
	int new_desc;
	dma_addr_t cont_addr;
	dma_addr_t addr;
	unsigned int cont_len;
	unsigned int len;
	unsigned int cont_max_len = 0;
	struct scatterlist *sgl;

	BUG_ON(!transfer);

	sgl = transfer->sgm->sgl;

	/* start first contiguous block */
	cont_addr = addr = sg_dma_address(&transfer->sgm->sgl[i]);
	cont_len = 0;

	/* iterate over all remaining entries but the last */
	for (i = 0; i < transfer->sgl_nents - 1; i++) {
		/* bus address of next entry i + 1 */
		dma_addr_t next = sg_dma_address(&sgl[i + 1]);
		/* length of this entry i */
		len = sg_dma_len(&sgl[i]);
		dbg_desc("SGLE %04d: addr=0x%016llx length=0x%08x\n", i,
			(u64)addr, len);

		/* add entry i to current contiguous block length */
		cont_len += len;

		new_desc = 0;
		/* entry i + 1 is non-contiguous with entry i? */
		if (next != addr + len) {
			dbg_desc("NON-CONTIGUOUS WITH DESC %d\n", i + 1);
			new_desc = 1;
		}
		/* entry i reached maximum transfer size? */
		else if (cont_len > (XDMA_DESC_MAX_BYTES - PAGE_SIZE)) {
			dbg_desc("BREAK\n");
			new_desc = 1;
		}

		if ((force_new_desc) && !(userspace))
			new_desc = 1;

		if (new_desc) {
			/* fill in descriptor entry j with transfer details */
			xdma_desc_set(transfer->desc_virt + j, cont_addr,
					ep_addr, cont_len, dir_to_dev);

			xdma_desc_force_complete(transfer->desc_virt + j);

			if (cont_len > cont_max_len)
				cont_max_len = cont_len;

			dbg_desc("DESC %4d:cont_addr=0x%llx\n", j,
				(u64)cont_addr);
			dbg_desc("DESC %4d:cont_len=0x%08x\n", j, cont_len);
			dbg_desc("DESC %4d:ep_addr=0x%llx\n", j, (u64)ep_addr);
			/* proceed EP address for next contiguous block */

			/* for non-inc-add mode don't increment ep_addr */
			if (userspace) {
				if (non_incr_addr == 0)
					ep_addr += cont_len;
			} else {
				ep_addr += cont_len;
			}

			/* start new contiguous block */
			cont_addr = next;
			cont_len = 0;
			j++;
		}
		/* goto entry i + 1 */
		addr = next;
	}
	/* i is the last entry in the scatterlist, add it to the last block */
	len = sg_dma_len(&sgl[i]);
	cont_len += len;
	BUG_ON(j > transfer->sgl_nents);

	/* j is the index of the last descriptor */

	dbg_desc("SGLE %4d: addr=0x%016llx length=0x%08x\n", i, (u64)addr, len);
	dbg_desc("DESC %4d: cont_addr=0x%llx cont_len=0x%08x ep_addr=0x%llx\n",
		j, (u64)cont_addr, cont_len, (unsigned long long)ep_addr);

	/* XXX to test error condition, set cont_len = 0 */

	/* fill in last descriptor entry j with transfer details */
	xdma_desc_set(transfer->desc_virt + j, cont_addr, ep_addr, cont_len,
		dir_to_dev);

	return j;
}

static struct xdma_transfer *transfer_create(struct xdma_dev *lro,
		const char *start, size_t cnt, u64 ep_addr, int dir_to_dev,
		int non_incr_addr, int force_new_desc, int userspace)
{
	int i = 0;
	int last = 0;
	int rc;
	struct scatterlist *sgl;
	struct xdma_transfer *transfer;
	u32 control;

	/* allocate transfer data structure */
	transfer = kzalloc(sizeof(struct xdma_transfer), GFP_KERNEL);

	dbg_sg("transfer_create()\n");

	if (!transfer)
		return NULL;

	/* remember direction of transfer */
	transfer->dir_to_dev = dir_to_dev;

	/* create virtual memory mapper */
	transfer->sgm = sg_create_mapper(cnt);
	BUG_ON(!transfer->sgm);
	transfer->userspace = userspace;

	/* lock user pages in memory and create a scatter gather list */
	if (userspace)
		rc = sgm_get_user_pages(transfer->sgm, start, cnt, !dir_to_dev);
	else
		rc = sgm_kernel_pages(transfer->sgm, start, cnt, !dir_to_dev);

	BUG_ON(rc < 0);

	sgl = transfer->sgm->sgl;

	dbg_sg("mapped_pages=%d.\n", transfer->sgm->mapped_pages);
	dbg_sg("sgl = 0x%p.\n", transfer->sgm->sgl);
	BUG_ON(!lro->pci_dev);
	BUG_ON(!transfer->sgm->sgl);
	BUG_ON(!transfer->sgm->mapped_pages);
	/* map all SG entries into DMA memory */
	transfer->sgl_nents = pci_map_sg(lro->pci_dev, transfer->sgm->sgl,
		transfer->sgm->mapped_pages,
		dir_to_dev ? DMA_TO_DEVICE : DMA_FROM_DEVICE);
	dbg_sg("hwnents=%d.\n", transfer->sgl_nents);

	/* verify if the page start address got into the first sg entry */
	dbg_sg("sg_page(&sgl[0])=0x%p.\n", sg_page(&transfer->sgm->sgl[0]));
	dbg_sg("sg_dma_address(&sgl[0])=0x%016llx.\n",
		(u64)sg_dma_address(&transfer->sgm->sgl[0]));
	dbg_sg("sg_dma_len(&sgl[0])=0x%08x.\n",
		sg_dma_len(&transfer->sgm->sgl[0]));

	/* allocate descriptor list */
	transfer->desc_virt = xdma_desc_alloc(lro->pci_dev, transfer->sgl_nents,
		&transfer->desc_bus, NULL);
	dbg_sg("transfer_create():\n");
	dbg_sg("transfer->desc_bus = 0x%llx.\n", (u64)transfer->desc_bus);

	last = transfer_build(transfer, ep_addr, dir_to_dev, non_incr_addr,
		force_new_desc, userspace);

	if (userspace) {
		transfer_perf(transfer, last);
	} else {
		/* terminate last descriptor */
		xdma_desc_link(transfer->desc_virt + last, 0, 0);
		/* stop engine, EOP for AXI ST, req IRQ on last descriptor */
		control = XDMA_DESC_STOPPED;
		control |= XDMA_DESC_EOP;
		control |= XDMA_DESC_COMPLETED;
		xdma_desc_control(transfer->desc_virt + last, control);
	}

	last++;
	/* last is the number of descriptors */
	transfer->desc_num = transfer->desc_adjacent = last;

	dbg_sg("transfer 0x%p has %d descriptors\n", transfer,
		transfer->desc_num);
	/* fill in adjacent numbers */
	for (i = 0; i < transfer->desc_num; i++) {
		xdma_desc_adjacent(transfer->desc_virt + i,
			transfer->desc_num - i - 1);
	}

	/* initialize wait queue */
	init_waitqueue_head(&transfer->wq);

	return transfer;
}

static int check_transfer_align(struct xdma_engine *engine,
	const char __user *buf, size_t count, loff_t pos, int sync)
{
	BUG_ON(!engine);

	/* AXI ST or AXI MM non-incremental addressing mode? */
	if (engine->streaming || engine->non_incr_addr) {
		int buf_lsb = (int)((uintptr_t)buf) & (engine->addr_align - 1);
		size_t len_lsb = count & ((size_t)engine->len_granularity - 1);
		int pos_lsb = (int)pos & (engine->addr_align - 1);

		dbg_tfr("AXI ST or MM non-incremental\n");
		dbg_tfr("buf_lsb = %d, pos_lsb = %d, len_lsb = %ld\n", buf_lsb,
			pos_lsb, len_lsb);

		if (buf_lsb != 0) {
			dbg_tfr("FAIL: non-aligned buffer address %p\n", buf);
			return -EINVAL;
		}

		if ((!engine->streaming) && (pos_lsb != 0) && (sync)) {
			dbg_tfr("FAIL: non-aligned AXI MM FPGA addr 0x%llx\n",
				(unsigned long long)pos);
			return -EINVAL;
		}

		if (len_lsb != 0) {
			dbg_tfr("FAIL: len %d is not a multiple of %d\n",
				(int)count,
				(int)engine->len_granularity);
			return -EINVAL;
		}
		/* AXI MM incremental addressing mode */
	} else {
		int buf_lsb = (int)((uintptr_t)buf) & (engine->addr_align - 1);
		int pos_lsb = (int)pos & (engine->addr_align - 1);

		if (buf_lsb != pos_lsb) {
			dbg_tfr("FAIL: Misalignment error\n");
			dbg_tfr("host addr %p, FPGA addr 0x%llx\n", buf, pos);
			return -EINVAL;
		}
	}

	return 0;
}

#endif

#define FLIP_CHANNEL 	FALSE	//TRUE will start using Read/Write channel 2 for testing only. Dafault should be FALSE

/* char_dma_read_write() -- Read from or write to the device
 *
 * @buf userspace buffer
 * @count number of bytes in the userspace buffer
 * @pos byte-address in device
 * @dir_to_device If !0, a write to the device is performed
 *
 * Iterate over the userspace buffer, taking at most 255 * PAGE_SIZE bytes for
 * each DMA transfer.
 *
 * For each transfer, get the user pages, build a sglist, map, build a
 * descriptor table. submit the transfer. wait for the interrupt handler
 * to wake us on completion.
 */

static ssize_t char_dma_read_write(struct file *filp, char __user *buf,
		size_t count, loff_t *pos, int dir_to_dev)
{

//dir_to_dev = 0 is write command and 1 is read command.
	/* fetch device specific data stored earlier during open */
	struct pedd_dev *priv = filp->private_data;
	UINT8 chan = 0;	//hardcoding to get first Chan0 or Read/Write working.
	if(FLIP_CHANNEL == TRUE)
		chan = 1;
	UINT8 wei = 0;
	BOOLEAN queue = 0;
	BOOLEAN lle = 0;

	PHYADDR src, dst;
	dma_addr_t *dma_bus = {{0} };	//DMA bus address
	enum dma_data_direction host_dma_dir = DMA_NONE; //DMA_BIDIRECTIONAL = 0, DMA_TO_DEVICE = 1, DMA_FROM_DEVICE = 2, DMA_NONE = 3.

	PHYADDR ppaddr = {{0} };	//PCI physical address
	PHYADDR dpaddr = {{0} };	//DDR physical address on device memory BAR(4) -> base + offsets in 4MB file

	if(priv == NULL)
		return -EINVAL;
	if((buf == NULL) || (count > MAX_SIZE_RW_ALLOWED))	//Reject if read/write is greater than 4K (that's the max sixe of contigious buffer/page available for use
		return -EINVAL;		//Remove erroring on count > 4K after SGL is supported.



	UINT32 status = STATUS_SUCCESS;
	/* default channel status is stopped */
	DMA_CHAN_STATUS stat = CHAN_STOPPED;
	HAL_DMA_RET ret;
	PDMA_RSC pdrsc = 	&priv->pedd_ext.DmaRsc ;//&pDevExt->DmaRsc;
	if(pdrsc == NULL)
	{
		printk(KERN_ERR "&priv->pedd_ext.DmaRsc = 0x%llx can't be null", pdrsc);
		return -EINVAL;
	}
	DMA_DIRC dirc = dir_to_dev == DIRECTION_TO_DEVICE_READ ? DMA_WRITE : DMA_READ;

	if ((dirc == DMA_WRITE && chan >= pdrsc->WrChans) || (dirc == DMA_READ && chan >= pdrsc->RdChans))
	{
		printk(KERN_ERR "dirc = %lu, chan = %lu, pdrsc->RdChans, pdrsc->WrChans\n", dirc, chan, pdrsc->RdChans, pdrsc->WrChans);
		return -EINVAL;

	}
	/* Check channel state to make sure it's not running, otherwise we can't
	* start a new transfer.
	*/
	ret = HalDmaState(pdrsc, dirc, chan, &stat, NULL, NULL, NULL, NULL);
	if (ret != DMA_SUCCESS) {
		printk(KERN_ERR "Failed to get DMA channel state:\n%s\n", HalDmaGetErrInfo(ret));
		status = STATUS_UNSUCCESSFUL;
		return -EINVAL;
	}
	if (stat == CHAN_RUNNING) {
		static char buf[2048];
		printk(KERN_ERR "DMA %s channel %d is still running.\n", dirc == DMA_WRITE ? "read command - DMA_WRITE" : "write command - DMA_READ", chan);
		status = STATUS_UNSUCCESSFUL;
		DmaDumpReg(pdrsc->DmaRegs, dirc, chan, buf, 2048);
		printk(KERN_ERR "%s", buf);
		/* Stop channel */
		DmaSetOp(pdrsc->DmaRegs, dirc, FALSE);
		/* use invalid mode, just to clear the register */
	    DmaSetCtrl(pdrsc->DmaRegs, dirc, chan, 2);
		return -EINVAL;
	}

	//First set correct channel number. 0 through 7 are Write and 8 through 15 are read. SO if we read on channel 0 then it's chan = chan0+8
	if (dirc == DMA_READ) //write command
		//First 8 buffers are for write channels (it's a DMA READ from device, so channel 0 is channel 8.
		ppaddr = priv->pedd_ext.IoMemPa[chan + 8]; //First 8 buffers are for write channels
	else	//read command
		//use perallocated Kernel memory buffer for DMA buffers.
		ppaddr = priv->pedd_ext.IoMemPa[chan]; //First 8 buffers are for write channels

	 //dpaddr = priv->pedd_ext.RamPa;
	 dpaddr.HighPart = 0x00;
	 //dpaddr.LowPart = DDR_BASE_ADDRESS_FOR_DMA; //Akida DDR address range.
	 //dpaddr.LowPart += *pos;
	 dpaddr.LowPart = *pos;

	if (dir_to_dev == DIRECTION_TO_DEVICE_READ) { //Read command - which is DMA WRITE from AKida DDR to Host memory buffer.
		//DIRECTION_TO_DEVICE_READ is DMA_WRITE on device side DMA
		host_dma_dir = DMA_FROM_DEVICE;
		dst =  ppaddr; //dma_bus;
		src =  dpaddr; //dma_bus;
	}
	else { //write command - which is DMA READ from  Host memory buffer to AKida DDR buffer.
		//DIRECTION_TO_DEVICE_WRITE
		host_dma_dir = DMA_TO_DEVICE;
		src =  ppaddr; //dma_bus;
		dst =  dpaddr; //dma_bus;
	}


//We are all set to start data transfer. NO SGL suport yet. Build SGL above this line if we have to support SGL.
	if(dir_to_dev == DIRECTION_TO_DEVICE_WRITE){
		//Write from host user buffer to device DDR destination
		//DIRECTION_TO_DEVICE_WRITE - write command
		if(lle == 0) {	//single buffer case
			//sync physical address for cpu access. Make sure it's not cached address
			pci_dma_sync_single_for_cpu(priv->pdev, (dma_addr_t)(priv->pedd_ext.IoMemPa[chan + 8].QuadPart), count, host_dma_dir);
			//dma_sync_virt(buf, count, host_dma_dir);

			if(copy_from_user(priv->pedd_ext.IoMemVa[chan + 8], buf, count) != 0)
			{
				printk(KERN_ERR "Write cmd Error copying from priv->pedd_ext.IoMemVa[chan + 8] = %lx to user buf = 0xlx, count = %lu\n", priv->pedd_ext.IoMemVa[chan + 8], buf, count );
				return -EIO;
			}
			DTRACE("Calling HalDmaSingleBlockRead for Write command src= 0x%llx, dst = 0x%llx, chan = %lu\n", src, dst, chan);
			//sync physical address for device DMA access. Flush any cache data to physical memory before starting transfer.
			pci_dma_sync_single_for_device(priv->pdev, (dma_addr_t)(priv->pedd_ext.IoMemPa[chan + 8].QuadPart), count, host_dma_dir);
			//make sure data in physical memory before transfer starts
			ret = HalDmaSingleBlockRead(&priv->pedd_ext.DmaRsc, chan, src, dst, count, wei, queue);
		}
		else { 	//SGL use case - TODO
			ret = DMA_NO_LL;             /**< No linked list are created */
			//ret = HalDmaMultiBlockRead(&priv->pedd_ext.DmaRsc, chan, src, dst, count, wei, queue);
		}
	}
	else { //it's a DIRECTION_TO_DEVICE_READ - read command
		//Read from Device DDR destination to host user buffer
		//DIRECTION_TO_DEVICE_READ
		if(lle == 0) {	 //single buffer case

			DTRACE("calling HalDmaSingleBlockWrite for Read command src= 0x%llx, dst = 0x%llx, chan = %lu, count = %d\n", src, dst, chan, count);
			ret = HalDmaSingleBlockWrite(&priv->pedd_ext.DmaRsc, chan, src, dst, count, wei, queue);
			//dma_sync_phys(priv->pedd_ext.IoMemPa[chan], count, host_dma_dir);
		}
		else {	//SGL use case - TODO
			ret = DMA_NO_LL;             /**< No linked list are created */
			//ret = HalDmaMultiBlockWrite(&priv->pedd_ext.DmaRsc, chan, src, dst, count, wei, queue);
		}
	}

	/* Transfer request is submitted, and we need to check the return value to
	* see if everything goes fine.
	*/
	if (ret == DMA_SUCCESS) {
		DTRACE("DMA_SUCCESS returned, now check channel status\n");
		uint32_t interruptCnt = 0;
		uint32_t watchDogCnt = 0;
		bool interruptRcvd = false;
		volatile uint64_t to = get_jiffies_64() + 500;
		while (!interruptRcvd) {
		    stat = HalDmaState(pdrsc, dirc, chan, &stat, NULL, NULL, &interruptCnt, NULL);
		    if(stat == DMA_SUCCESS) {
		        if (interruptCnt) {
			    	interruptRcvd = true;
				}
				if (time_is_before_jiffies64(to)) {
					interruptRcvd = true; // let it through so as to not hang kernel process
					status = STATUS_UNSUCCESSFUL;
					printk(KERN_INFO "Timeout %llu %llu reset !\n", jiffies, to);
					DmaSetOp(pdrsc->DmaRegs, dirc, FALSE);
					DmaSetCtrl(pdrsc->DmaRegs, dirc, chan, 2/*invalid, to clear the reg*/);
				}
			}
		}

		if(dir_to_dev == DIRECTION_TO_DEVICE_READ) {
	        if(stat == DMA_SUCCESS) {
	        	//sync physical address for cpu access. Make sure it's not cached address
	        	pci_dma_sync_single_for_cpu(priv->pdev, (dma_addr_t)(priv->pedd_ext.IoMemPa[chan].QuadPart), count, host_dma_dir);
	        	if(copy_to_user(buf, priv->pedd_ext.IoMemVa[chan], count) != 0)
	        	{
	        		printk(KERN_ERR "Read cmd Error copying from priv->pedd_ext.IoMemVa[chan] = %llx to user buf = %xllx, count = %lu\n", priv->pedd_ext.IoMemVa[chan], buf, count);
	        		return -EIO;
	        	}
	        	DTRACE("Read cmd copying from priv->pedd_ext.IoMemVa[chan] = 0x%lx to user buf = 0x%lx, count = %d\n", priv->pedd_ext.IoMemVa[chan], buf, count);
	        	//dma_sync_virt(buf, count, host_dma_dir);
	        }
	        else
	        {
	        	printk(KERN_ERR "Read transfer not completed or timed out! after 5 sec\n");
	        	return -EIO;
	        }
		}
		else {
			if(stat == DMA_SUCCESS) {
				DTRACE("write transfer completed\n");
			}
			else {
				printk(KERN_ERR "Write transfer not completed or timed out! after 5 sec\n");
				return -EIO;
			}
        }
	}
	else {
		printk(KERN_ERR "Failed to start DMA %s:\n%s\n", dir_to_dev == DIRECTION_TO_DEVICE_WRITE ? "write" : "read", HalDmaGetErrInfo(ret));
		status = STATUS_UNSUCCESSFUL;
	}


#if 0	//Debug only
	//print data from kernel buffer for debuging
	unsigned int ii = 0;
	for(ii = 0; ii < 20;ii++)
	{
		PUINT32 pvaddr;
		//use perallocated Kernel memory buffer for DMA buffers.
		if (dirc == DMA_READ)
		{
			pvaddr = priv->pedd_ext.IoMemVa[chan + 8]; //First 8 buffers are for write channels
		}
		else
		{
			pvaddr = priv->pedd_ext.IoMemVa[chan]; //First 8 buffers are for write channels
		}
		DTRACE("RW buffer at %lx = %lx", (pvaddr + ii), *(pvaddr + ii) );
	}

#endif

	if (status == 0) {
		/* if transfer was successful, update position */
		*pos += count;
		return count;
	}
	return status;
}


loff_t pedd_llseek(struct file *filp, loff_t off, int whence)
{
	/* fetch device specific data stored earlier during open */
	struct pedd_dev *priv = filp->private_data;
	loff_t newpos;

	if(priv == NULL)
		return -EINVAL;

	DTRACE("Seeking to DDR Buffer offset 0x%llx, whence = %ul\n", off, whence);
	switch(whence)
	{
	 	   case 0: /* SEEK_SET */
	   		//File size if 4MB of the DDR Buffer in BAR4. Make sure it does not seek beyond 4MB.
			   #if 0
	   		   if(off >= AKIDA_DMA_TO_DDR_TOTAL_SIZE) {
	                           DTRACE("pedd_llseek: error off is > than DDR SIZE");
	   			   return -EINVAL;
                           }
	   		   else
	   			   newpos = off;
                           #endif
	   		   newpos = off;
	   		   break;

	   	   case 1: /* SEEK_CUR */
	   		   if((filp->f_pos + off) >= AKIDA_DMA_TO_DDR_TOTAL_SIZE)
	   			   return -EINVAL;
	   		   else
	   			   newpos = filp->f_pos + off;
	   		   break;

	   	   case 2: /* SEEK_END */
	   		//filp->f_size = priv->bar_len[4]; //Size of DDR buffer is max file size.
	   		//seek to offset from end of device. This is just for doing DMA beyond 4MB but not really safe. Experimental.
	   		   newpos = priv->bar_len[4] + off;
	   		   break;

	   	   default: /* can't happen */
	   		   return -EINVAL;
	  }
#if 0
	  if (newpos < 0) {
	          DTRACE("pedd_llseek: error newpos is < 0");
		  return -EINVAL;
          }
#endif

	  DTRACE("Setting to offset 0x%llx, whence = %ul\n", off, whence);
	  filp->f_pos = newpos;
	  return newpos;
}

static ssize_t pedd_read(struct file *filp, char __user *buf, size_t count,
			 loff_t *f_pos)
{
	//Check for  valid parameters for Read DMA. DMA from Device memory to Host memory. dir_to_dev = 0
	DTRACE("Called Pedd_Read - address 0x%8x, count %d, pos 0x%8x\n", buf, count, f_pos);
	return char_dma_read_write(filp, (char *)buf, count, f_pos, DIRECTION_TO_DEVICE_READ);

}

static ssize_t pedd_write(struct file *filp, const char __user *buf,
			  size_t count, loff_t *f_pos)
{
	DTRACE("Called Pedd_Write - address 0x%8x, count %d, pos 0x%8x\n", buf, count, f_pos);

	//Check for  valid parameters for write DMA. DMA from Host memory to device memory. dir_to_dev = 1
	return char_dma_read_write(filp, (char *)buf, count, f_pos, DIRECTION_TO_DEVICE_WRITE);

}

int akida_regs_mmap(struct file *filp, struct vm_area_struct *vma)
{
	struct pedd_dev *priv = filp->private_data;

	unsigned long off;
	unsigned long phys; //void * __iomem
	unsigned long vsize;
	unsigned long psize;
	int rv;



	off = vma->vm_pgoff << PAGE_SHIFT;
	vsize = vma->vm_end - vma->vm_start;

	/* complete resource */
	psize = priv->mmio_len - off;
	//Make sure mmap is not mapping bigger than bar 0 size
	if (vsize > psize)
		return -EINVAL;

	/* BAR physical address */
	//Calculate offset with PCIe Bar0 start address
	phys = (char *)priv->mmio_start + off;

	DTRACE("mmap AKida Bar0 start phys = 0x%llx, size = %lu\n", (unsigned long long) phys, psize);
	DTRACE("mmap vm->vm_pgoff = ox%lx,vsize = %lu\n", vma->vm_pgoff, vsize);

	/*
	 * pages must not be cached as this would result in cache line sized
	 * accesses to the end point
	 */
	vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);
	/*
	 * prevent touching the pages (byte access) for swap-in,
	 * and prevent the pages from being swapped out. Lock it as it's akida registers used for VM I/O.
	 */
	vma->vm_flags |= VMEM_FLAGS;
	/* make bar[0] address MMIO accessible to user space */
	rv = io_remap_pfn_range(vma, vma->vm_start, phys >> PAGE_SHIFT, vsize, vma->vm_page_prot);
	if (rv)
		return -EAGAIN;
	else
		DTRACE("vma=0x%p, vma->vm_start=0x%lx, phys=0x%lx, vsize=%lu, ret-value = %u\n",
				vma, vma->vm_start, phys, vsize, rv);
	return 0;
}

static long pedd_ioctl(struct file *filp, unsigned int cmd, unsigned long arg)
{
	struct pedd_dev *priv = filp->private_data;
	int ret = 0;
	char ping_str[20] = "hello from pedd";
	static CHAR msg[MAX_MSG_SIZE];
	CMD_INFO cmdinfo;

	PciCfg_GetHeader_USER_OUT       pci_hdr;

	Pci_Header_Read  hdr_read;
	Pci_Header_Write hdr_write;
	Pci_Bar_Read     bar_read;
	Pci_Bar_Write    bar_write;

	UINT32 num, xferSize, status, bar;
	int rc;

	INFO_PRINT("pedd_ioctl Enter, cmd = 0x%x\n", cmd);

	switch (cmd) {
	case PED_IOCTL_PING:
		INFO_PRINT("PED_IOCTL_PING\n");
		ret = copy_to_user((char __user *)arg, ping_str, strlen(ping_str)+1);
		break;
	case PED_IOCTL_PCI_HDR_READ:
		INFO_PRINT("PED_IOCTL_PCI_HDR_READ\n");
		pci_hdr.StatusStr[0] = '\0';
		pedd_Read_Header(priv->pdev, &pci_hdr.pciData);
		pci_hdr.status = STATUS_SUCCESS;
		// Send data to user space
		ret = copy_to_user((PciCfg_GetHeader_USER_OUT __user *)arg, &pci_hdr, sizeof(PciCfg_GetHeader_USER_OUT));
		break;
	case PED_IOCTL_PCI_HDR_DATA_READ:
		INFO_PRINT("PED_IOCTL_PCI_HDR_DATA_READ\n");
		ret = copy_from_user(&hdr_read, (Pci_Header_Read __user *)arg, sizeof(Pci_Header_Read));
		if (ret) {
			ERROR_PRINT("copy_from_user fail\n");
			return ret;
		}

		num = (hdr_read.ioctl_in.Num != 0) ? hdr_read.ioctl_in.Num : 1;
		INFO_PRINT("DATA_READ offset=0x%lx len=%lu num=%d\n", hdr_read.ioctl_in.Offset, hdr_read.ioctl_in.Length, num);

		hdr_read.ioctl_out.StatusStr[0] = '\0';
		rc = hdr_read.ioctl_in.Rc;
		ret = pedd_rw_pci_hdr(priv->pdev, rc, 0, hdr_read.ioctl_in.Length, hdr_read.ioctl_in.Offset, &(hdr_read.ioctl_out.Data[0]));
		hdr_read.ioctl_out.status = (ULONG)ret;
		if (ret == STATUS_UNSUCCESSFUL)
			strcpy(hdr_read.ioctl_out.StatusStr, "Can't find snps's RC\n");
		ret = copy_to_user((Pci_Header_Read __user *)arg, &hdr_read, sizeof(Pci_Header_Read));
		break;
	case PED_IOCTL_PCI_HDR_DATA_WRITE:
		INFO_PRINT("PED_IOCTL_PCI_HDR_DATA_WRITE\n");
		ret = copy_from_user(&hdr_write, (Pci_Header_Write __user *)arg, sizeof(Pci_Header_Write));
		if (ret) {
			ERROR_PRINT("copy_from_user fail\n");
			return ret;
		}

		num = (hdr_write.ioctl_in.Num != 0) ? hdr_write.ioctl_in.Num : 1;
		INFO_PRINT("DATA_WRITE offset=0x%lx len=%lu num=%d\n", hdr_write.ioctl_in.Offset, hdr_write.ioctl_in.Length, num);

		hdr_write.ioctl_out.StatusStr[0] = '\0';
		rc = hdr_write.ioctl_in.Rc;
		ret = pedd_rw_pci_hdr(priv->pdev, rc, 1, hdr_write.ioctl_in.Length, hdr_write.ioctl_in.Offset, &(hdr_write.ioctl_in.Data[0]));
		hdr_write.ioctl_out.status = (ULONG)ret;
		if (ret == STATUS_UNSUCCESSFUL)
			strcpy(hdr_write.ioctl_out.StatusStr, "Can't find snps's RC\n");

		ret = copy_to_user((Pci_Header_Write __user *)arg, &hdr_write, sizeof(Pci_Header_Write));
		break;
	case PED_IOCTL_PCI_BAR_READ:
		INFO_PRINT("PED_IOCTL_PCI_BAR_READ\n");
		// Get request from user space
		ret = copy_from_user(&bar_read, (Pci_Bar_Read __user *)arg, sizeof(Pci_Bar_Read));
		if (ret) {
			ERROR_PRINT("copy_from_user fail\n");
			return ret;
		}

		bar = bar_read.ioctl_in.Bar;
		if (!priv->bar[bar]) {
			ERROR_PRINT("BAR #%d is not present, can't read\n", bar);
			return -ENOTTY;
		}
		num = (bar_read.ioctl_in.Num != 0) ? bar_read.ioctl_in.Num : 1;
		num = (num > 8) ? 8 : num;

		INFO_PRINT("BAR READ bar=%lu offset=0x%lx len=%lu num=%d",
		bar_read.ioctl_in.Bar, bar_read.ioctl_in.Offset, bar_read.ioctl_in.Length, num);

		bar_read.ioctl_out.StatusStr[0] = '\0';
		pedd_rw_pci_bar(0, bar_read.ioctl_in.Length, priv->bar[bar], bar_read.ioctl_in.Offset, &(bar_read.ioctl_out.Data[0]));
		bar_read.ioctl_out.status = STATUS_SUCCESS;

		ret = copy_to_user((Pci_Bar_Read __user *)arg, &bar_read, sizeof(Pci_Bar_Read));
		break;
	case PED_IOCTL_PCI_BRA_WRITE:
		INFO_PRINT("PED_IOCTL_PCI_BRA_WRITE\n");
		ret = copy_from_user(&bar_write, (Pci_Bar_Write __user *)arg, sizeof(Pci_Bar_Write));
		if (ret) {
			ERROR_PRINT("copy_from_user fail\n");
			return ret;
		}

		bar = bar_write.ioctl_in.Bar;
		if (!priv->bar[bar]) {
			ERROR_PRINT("BAR #%d is not present, can't write\n", bar);
			return -ENOTTY;
		}
		num = (bar_write.ioctl_in.Num != 0) ? bar_write.ioctl_in.Num : 1;
		num = (num > 8) ? 8 : num;

		INFO_PRINT("BAR WRITE offset=0x%lx len=%lu num=%d\n", bar_write.ioctl_in.Offset, bar_write.ioctl_in.Length, num);

		bar_write.ioctl_out.StatusStr[0] = '\0';
		pedd_rw_pci_bar(1, bar_write.ioctl_in.Length, priv->bar[bar], bar_write.ioctl_in.Offset, &(bar_write.ioctl_in.Data[0]));
		bar_write.ioctl_out.status = STATUS_SUCCESS;

		ret = copy_to_user((Pci_Bar_Write __user *)arg, &bar_write, sizeof(Pci_Bar_Write));
		break;
	case PED_IOCTL_APP_CMD:
		xferSize = strnlen_user((char __user *)arg, MAX_MSG_SIZE);
		ret = strncpy_from_user(msg, (char __user *)arg, (xferSize + 1));
		msg[(xferSize >= MAX_MSG_SIZE) ? MAX_MSG_SIZE - 1:xferSize] = '\0';

		/* Parse the request from user space */
		if (ParseCommand(msg, &cmdinfo)) {
			/* This is the main function to handle all DMA related commands. */
			status = IoctlCommand(priv, &priv->pedd_ext, &cmdinfo);
			ret = copy_to_user((char __user *)arg, cmdinfo.ResMsg, strlen(cmdinfo.ResMsg) + 1);
		} else {
			ERROR_PRINT("PEDD_IOCTL_APP_CMD - [cmd=%s] PARSE FAILED\n", msg);
			ret = copy_to_user((char __user *)arg, cmdinfo.ResMsg, strlen(cmdinfo.ResMsg) + 1);
		}
		break;
	default:
		ERROR_PRINT("wrong ioctl cmd\n");
		return -ENOTTY;
	}

	return ret;
}

static int pedd_open(struct inode *inode, struct file *filp)
{
	struct pedd_dev *priv;

	INFO_PRINT("pedd_open Enter\n");

	priv = container_of(inode->i_cdev, struct pedd_dev, cdev);
	filp->private_data = priv;
#if 0
	if(priv->pedd_ext.ChunkLen == 0) //then DMA is not initialized
	{
		//Now initialize DMA.
		CMD_INFO *pcmd; //Just for holding some return messages
		pcmd->Cmd = DMA_CMD_INIT;
		//use ChunkLen as a signature to check if DMA is already initialized.
		priv->pedd_ext.ChunkLen = 0;
		//Later check in commands to see if it is initialized to something > 0
		IoctlDmaInit(priv, &priv->pedd_ext, pcmd);
	}
	else
		DTRACE("DMA is already initialized during previous character device open\n");
#endif
	INFO_PRINT("pedd_open - finished DMA_INIT\n");
	return 0;
}

static int pedd_release(struct inode *inode, struct file *filp)
{
	INFO_PRINT("pedd release!\n");

	return 0;
}

static int pedd_cdev_init(struct pedd_dev *peddev)
{
	int ret;

	INFO_PRINT("pedd_cdev_init Enter\n");

	ret = alloc_chrdev_region(&pedd_devno, 0, pedd_nr_devs, "pedd");
	if (ret < 0) {
		ERROR_PRINT("can't get cdev major\n");
		return ret;
	}

	pedd_cdev_major = MAJOR(pedd_devno);
	pedd_devno = MKDEV(pedd_cdev_major, pedd_cdev_minor);
	cdev_init(&peddev->cdev, &pedd_fops);
	peddev->cdev.owner = THIS_MODULE;
	peddev->cdev.ops = &pedd_fops;

	ret = cdev_add(&peddev->cdev, pedd_devno, 1);
	peddev->devno = pedd_devno;
	if (ret < 0)
		ERROR_PRINT("can not add pedd cdev\n");

	return ret;

}

static void pedd_priv_init(struct pedd_dev *priv)
{
	PDEVICE_EXTENSION pDevExt;
	UINT32 dcr;

	pDevExt = &priv->pedd_ext;
	tasklet_init(&priv->tlet, DataCheckDpc, (unsigned long)priv);

	/* initialize the pedd data
	 */

	RtlZeroMemory(&pDevExt->DmaRsc, sizeof(DMA_RSC));
	RtlZeroMemory(&pDevExt->xDataRsc, sizeof(XDATA_RSC));

	/* Get DMA registers */
	if (eDMA || xData) {
		void __iomem *tmp = (void __iomem *) priv->bar[eDMA_regs_bar];
		tmp += eDMA_regs_off;
		pDevExt->DmaRsc.DmaRegs = (PDMA_REGS)tmp;
	}

	/* Get local RAM addresses
	 * Note: in latest prototype (02/15/2014), the onboard RAM space is
	 * changed. It's splitted to two parts, and we only use the second part.
	 */
	if (eDMA || xData) {
		pDevExt->RamSize = (UINT32)priv->bar_len[eDMA_mem_bar];
		// Just a restriction, because the maximum RAM size that eDMA would even use is 64MB
		if (pDevExt->RamSize > (64 * 1024 * 1024))
			pDevExt->RamSize = 64 * 1024 * 1024;
		pDevExt->RamVa = (PUINT32)priv->bar[eDMA_mem_bar];
		pDevExt->RamPa.QuadPart = priv->RegSpacePa[eDMA_mem_bar].QuadPart;
	}

	/* Get local xData addresses */
	if (xData)
		pDevExt->xDataRsc.xDataRegs = priv->bar[xData_bar];

#if 0
	/* Get interrupt object */
	if (pDEC->MessageInterruptCount == 1) {
		PIO_INTERRUPT_MESSAGE_INFO msgInfo =
				(PIO_INTERRUPT_MESSAGE_INFO) pDEC->IntConnectionContext;
		pDevExt->DmaRsc.IntrObj
				= msgInfo->MessageInfo[0].InterruptObject;
	} else {
		LWDM_ERROR(pDEC, "Failed to obtain MSI interrupt object");
		goto CLEANUP;
	}
#endif

	/* Reset other device extension members */
	RtlZeroMemory(&pDevExt->IoMemVa, 16 * sizeof(PUINT32));
	RtlZeroMemory(&pDevExt->IoMemPa, 16 * sizeof(PHYADDR));
	RtlZeroMemory(&pDevExt->WrChInfEx, 8 * sizeof(CHAN_INFO_EX));
	RtlZeroMemory(&pDevExt->RdChInfEx, 8 * sizeof(CHAN_INFO_EX));
	pDevExt->ChunkLen = 0;

	/* Max PCI packet payload size */
	pedd_rw_pci_cfg(priv->pdev, 0, 4, 0x78, (CHAR *)&dcr);

	pDevExt->WriteMTU = 128;
	pDevExt->WriteMTUBits = 0xFE000000;
	switch ((dcr >> 5) & 0x07) {
		case 0:
			pDevExt->WriteMTU = 128;
			pDevExt->WriteMTUBits = 0xFE000000;
			break;
		case 1:
			pDevExt->WriteMTU = 256;
			pDevExt->WriteMTUBits = 0xFF000000;
			break;
		case 2:
			pDevExt->WriteMTU = 512;
			pDevExt->WriteMTUBits = 0xFF800000;
			break;
		case 3:
			pDevExt->WriteMTU = 1024;
			pDevExt->WriteMTUBits = 0xFFC00000;
			break;
		case 4:
			pDevExt->WriteMTU = 2048;
			pDevExt->WriteMTUBits = 0xFFE00000;
			break;
		case 5:
			pDevExt->WriteMTU = 4096;
			pDevExt->WriteMTUBits = 0xFFF00000;
			break;
	}
	pDevExt->ReadMTU = 128;
	pDevExt->ReadMTUBits = 0xFE000000;
	switch ((dcr >> 12) & 0x07) {
		case 0:
			pDevExt->ReadMTU = 128;
			pDevExt->ReadMTUBits = 0xFE000000;
			break;
		case 1:
			pDevExt->ReadMTU = 256;
			pDevExt->ReadMTUBits = 0xFF000000;
			break;
		case 2:
			pDevExt->ReadMTU = 512;
			pDevExt->ReadMTUBits = 0xFF800000;
			break;
		case 3:
			pDevExt->ReadMTU = 1024;
			pDevExt->ReadMTUBits = 0xFFC00000;
			break;
		case 4:
			pDevExt->ReadMTU = 2048;
			pDevExt->ReadMTUBits = 0xFFE00000;
			break;
		case 5:
			pDevExt->ReadMTU = 4096;
			pDevExt->ReadMTUBits = 0xFFF00000;
			break;
	}

}

static void pedd_priv_exit(struct pedd_dev *priv)
{
	PDEVICE_EXTENSION pDevExt;
	int i;

	pDevExt = &priv->pedd_ext;

	/* Free eDMA system memory */
	for (i = 0; i < 16; i++)
		FreeProtocolMem(priv, pDevExt->ChunkLen * 4,
				&pDevExt->IoMemVa[i],
				&pDevExt->dma_bus[i]);

	/* Free xData memory */
	if (pDevExt->xDataRsc.xDataMemVa)
		FreeProtocolMem(priv, pDevExt->xDataRsc.xDataMemSize,
				&pDevExt->xDataRsc.xDataMemVa,
				&pDevExt->xDataRsc.xDataBus);
}

//#define AKIDA_IATU_OFFSET_IN_CONFIG_SPACE	0x900 //Consult Hussein for correct offset of iATU registers in Config space
#define USE_IATU	TRUE	//To experiment with iATU enable from driver. Dafault should be FALSE
int akida_bar_map_iATU(struct pci_dev *pdev)
{
	//int	pci_read_config_dword(const struct pci_dev *dev, int where, u32 *val)
	//int pci_write_config_word(const struct pci_dev *dev, int where, u16 val)
	int rc = -ENODEV;

        DTRACE("Start initializing Akida BAR's mapping by programming iATU\n");

	if(pdev != NULL)
	{

		 //Program iATU - AKida bar

		pci_write_config_dword(pdev, 0x0900, 0x80000000);  // Region 0 inbound (31:dir N-0:Region Index)
		pci_write_config_dword(pdev,0x0904, 0x00000000);  // Function 0 Mem   (22-20:function)
		pci_write_config_dword(pdev,0x0918, 0xFCC00000);  // Lower Target -  AKIDA CORE APB
		pci_write_config_dword(pdev,0x0908, 0xC0080000);  // CTRL2            (31:EN, 30:BAR Match, 19:FUNC EN, 10-8:BAR0)

	//----- EP_iATU Region 1 Inbound Setting -----//
		pci_write_config_dword(pdev,0x0900, 0x80000001);  // Region 1 inbound (31:dir N-0:Region Index)
		pci_write_config_dword(pdev,0x0904, 0x00000000);  // Function 0 Mem   (22-20:function)
		pci_write_config_dword(pdev,0x0918, 0xF8C00000);  // PCIe EP DBI APB System Address - DMA controller mapped here at offset 0x970
		pci_write_config_dword(pdev,0x0908, 0xC0080200);  // CTRL2            (31:EN, 30:BAR Match, 19:FUNC EN, 10-8:BAR2)

	//----- EP_iATU Region 2 Inbound Setting -----//
		pci_write_config_dword(pdev,0x0900, 0x80000002);  // Region 2 inbound (31:dir N-0:Region Index)
		pci_write_config_dword(pdev,0x0904, 0x00000000);  // Function 0 memory for testing DstBuffer in DDR at 4MB   (22-20:function)
		pci_write_config_dword(pdev,0x0918, 0x20000000);  // LPDDR APB System address - used for SGL - LL is first 1MB, rest 3MB is used for Data Buffers.
		pci_write_config_dword(pdev,0x0908, 0xC0080400);  // CTRL2            (31:EN, 30:BAR Match, 19:FUNC EN, 10-8:BAR4)

		//Setup bars successfully
		rc = 0;
	}
	//make sure to provide 100msec sleep for iATU's to be setup.
	KERNEL_WAIT(1000);

	return rc;

}

static int pedd_probe(struct pci_dev *pdev, const struct pci_device_id *pci_id)
{
	unsigned long mmio_flags;
	resource_size_t mmio_start, mmio_end, mmio_len;
	void *mmio_addr = NULL;
	struct pedd_dev *priv = NULL;
	int rc = -ENODEV;
	struct device *dev = &pdev->dev;

	DTRACE("pedd_probe: Enter\n");
	INFO_PRINT("eDMA %sabled by parameter\n", eDMA ? "en" : "dis");
	if (eDMA) {
#ifdef PCIE_UNROLL_SUPPORT
		INFO_PRINT("PED driver with DMA Unroll support\n");
#else
		INFO_PRINT("PED driver with DMA Legacy support\n");
#endif
		INFO_PRINT("eDMA mem BAR: %d\n", eDMA_mem_bar);
		INFO_PRINT("eDMA regs BAR: %d, offset 0x%.8lx\n", eDMA_regs_bar, eDMA_regs_off);
	}

	INFO_PRINT("xData %sabled by parameter\n", xData ? "en" : "dis");
	if (xData) {
		INFO_PRINT("xData BAR configured: %d\n", xData_bar);
	}

	INFO_PRINT("HW performance workaround %sabled by parameter\n", HW_performance_workaround ? "en" : "dis");

	priv = kzalloc(sizeof(struct pedd_dev), GFP_KERNEL);
	if (!priv)
		return -ENOMEM;

	priv->pdev = pdev;
	pci_set_drvdata(pdev, priv);

	//iATU code needs debug. don't enable defalt.
	if(USE_IATU == TRUE) {
		//Initialise AKIDA iATU BAR mappings before enabling bus mastering which will allow it to start reading/writing memory bar's
		//offset in config space is same as we see offset for a register. Say DMA registers at offset 0x970
		if(akida_bar_map_iATU(pdev) != 0)
		{
			ERROR_PRINT("%s Failed to Initialize Akida BAR's iATU. No device found\n", pci_name(pdev));
			goto err_out;
		}
	}
	/* enable device */
	if (pci_enable_device(pdev)) {
		ERROR_PRINT("%s cannot enable device\n", pci_name(pdev));
		goto err_out;
	}

	/* enable bus master capability on device */
	pci_set_master(pdev);

	if (MSI_only)
		rc = pci_alloc_irq_vectors(pdev, 1, 1, PCI_IRQ_MSI);
	else
		rc = pci_alloc_irq_vectors(pdev, 1, 1, PCI_IRQ_MSI | PCI_IRQ_MSIX);
	if (rc < 0) {
		INFO_PRINT("Could not enable interrupt\n");
		rc = -EPERM;
		goto err_out;
	}

	mmio_start = pci_resource_start(pdev, BAR_MEM);
	mmio_end = pci_resource_end(pdev, BAR_MEM);
	mmio_flags = pci_resource_flags(pdev, BAR_MEM);
	mmio_len = pci_resource_len(pdev, BAR_MEM);

	if (!(mmio_flags & IORESOURCE_MEM)) {
		ERROR_PRINT("cannot find proper PCI device base address\n");
		rc = -ENODEV;
		goto err_out;
	}

	rc = pci_request_regions(pdev, DRVER_NAME);
	if (rc) {
		priv->in_use = 1;
		goto err_out;
	}
	priv->got_regions = 1;

	mmio_addr = ioremap(mmio_start, mmio_len);
	//mmio_addr = ioremap_nocache(mmio_start, mmio_len);
	if (!mmio_addr) {
		ERROR_PRINT("%s:cannot remap mmio, aborting\n", pci_name(pdev));
		rc = -EIO;
		goto err_out;
	}
	INFO_PRINT("mmio_addr = 0x%p\n", (void *)mmio_addr);

	if (!pci_set_dma_mask(pdev, DMA_BIT_MASK(64))) {
		pci_set_consistent_dma_mask(pdev, DMA_BIT_MASK(64));
		INFO_PRINT("Using a 64-bit DMA mask\n");
	} else {
		if (!pci_set_dma_mask(pdev, DMA_BIT_MASK(32))) {
			pci_set_consistent_dma_mask(pdev, DMA_BIT_MASK(32));
			INFO_PRINT("Using a 32-bit DMA mask\n");
		} else {
			ERROR_PRINT("No suitable DMA possible\n");
			rc = -1;
			goto err_mask;
		}
	}

	pci_read_config_byte(pdev, PCI_REVISION_ID, &priv->revision);
	pci_read_config_byte(pdev, PCI_INTERRUPT_PIN, &priv->irq_pin);
	pci_read_config_byte(pdev, PCI_INTERRUPT_LINE, &priv->irq_line);
	pci_read_config_word(pdev, PCI_SUBSYSTEM_VENDOR_ID,
				   &priv->sub_vendor_id);
	pci_read_config_word(pdev, PCI_SUBSYSTEM_ID, &priv->sub_system_id);
	pci_read_config_word(pdev, PCI_VENDOR_ID, &priv->vendor_id);
	pci_read_config_word(pdev, PCI_DEVICE_ID, &priv->device_id);

	priv->mmio_addr = mmio_addr;
	priv->mmio_start = (void *)mmio_start;
	priv->mmio_len = mmio_len;
	priv->irq = pci_irq_vector(pdev, 0);	
	if (eDMA)
		priv->pedd_ext.DmaRsc.Irq = priv->irq;

	rc = devm_request_irq(dev, pci_irq_vector(pdev, 0), pedd_isr, 0, DRVER_NAME, (void *)priv);
	if (rc) {
		ERROR_PRINT("IRQ #%d, error %d\n", pci_irq_vector(pdev, 0), rc);
		goto err_irq;
	}

	pedd_map_bars(priv, pdev);

	rc = pedd_cdev_init(priv);
	if(rc < 0)
		goto err_cdev_init;

	pedd_priv_init(priv);

	priv->drv_cls = class_create(THIS_MODULE, DRVER_CLASS_NAME);
	if (IS_ERR(priv->drv_cls)) {
		ERROR_PRINT("failed in creating class\n");
		goto err_class_create;
	}

	device_create(priv->drv_cls, NULL, priv->devno, NULL, DRVER_NAME "%d",
		      MINOR(priv->devno));

	//Now initialize DMA.
	CMD_INFO *pcmd = kzalloc(sizeof(CMD_INFO), GFP_KERNEL); //Just for holding some return messages
	pcmd->Cmd = DMA_CMD_INIT;
	//use ChunkLen as a signature to check if DMA is already initialized.
	priv->pedd_ext.ChunkLen = 0;
	//Later check in commands to see if it is initialized to something > 0
	IoctlDmaInit(priv, &priv->pedd_ext, pcmd);

	printk(KERN_INFO "pedd_probe: done\n");
	return 0;

err_class_create:
err_cdev_init:
	devm_free_irq(dev, pci_irq_vector(pdev, 0), (void *)priv);
err_irq:
	pci_free_irq_vectors(pdev);
	/* disable the device if it is not in use */
	if (!priv->in_use)
		pci_disable_device(pdev);
	if (priv->got_regions)
		pci_release_regions(pdev);

err_mask:
err_out:
	kfree(priv);
	return rc;
}

static void pedd_remove(struct pci_dev *pdev)
{
	struct pedd_dev *priv;
	struct device *dev = &pdev->dev;

	priv = (struct pedd_dev *)pci_get_drvdata(pdev);

	cdev_del(&priv->cdev);
	unregister_chrdev_region(priv->devno, pedd_nr_devs);
	device_destroy(priv->drv_cls, priv->devno);
	class_destroy(priv->drv_cls);  //delete class created by us

	pedd_priv_exit(priv);

	if (pci_irq_vector(pdev, 0))
		devm_free_irq(dev, pci_irq_vector(pdev, 0), (void *)priv);
	pci_free_irq_vectors(pdev);

	pedd_unmap_bars(priv, pdev);

	iounmap(priv->mmio_addr);
	pci_release_regions(pdev);
	priv->pdev = NULL;
	kfree(priv);

	pci_set_drvdata(pdev, NULL);
	pci_disable_device(pdev);
	printk(KERN_INFO "pedd_remove: done\n");
}


static irqreturn_t pedd_isr(int irq, void *dev_id)
{
	PDMA_RSC pDmaRsc;
	struct pedd_dev *priv = (struct pedd_dev *)dev_id;

	if (!priv)
		return IRQ_NONE;

	if (eDMA) {
		pDmaRsc = &priv->pedd_ext.DmaRsc;

		/* Call HAL ISR to get interrupt source, clear interrupt,
		 * update linked list elements etc.
		 */
		HalDmaISR(pDmaRsc);
	}

	priv->irq_count++;

	return IRQ_HANDLED;
}

static int __init pedd_init_module(void)
{
	return pci_register_driver(&pedd_pci_driver);
}

static void __exit pedd_cleanup_module(void)
{
	pci_unregister_driver(&pedd_pci_driver);
}


module_init(pedd_init_module);
module_exit(pedd_cleanup_module);
